#!/usr/bin/env python

"""This module is used to define the radar network configuration.

TODO: wrong module description?

"""

import typing as t
import numpy as np
import numpy.typing as npt
import scipy.constants


def rayleigh_separatrix_range(
    gain_tx,
    gain_rx,
    wavelength,
    power_tx,
    snr,
    coherent_integration_time=0.02,
    effective_noise_temperature=150.0,
    radar_albedo=1.0,
):
    """
    The range at which the Optical-to-Rayleigh change occurs given the input parameters.
    """

    rx_noise = (
        scipy.constants.k * effective_noise_temperature / coherent_integration_time
    )
    power = snr * rx_noise / radar_albedo
    separatrix_diameter = wavelength / (np.pi * np.sqrt(3.0))
    r = (
        power_tx
        * gain_tx
        * gain_rx
        * wavelength**2.0
        * separatrix_diameter**2
        / (256.0 * np.pi**2 * power)
    ) ** (1.0 / 4.0)

    return r


def hard_target_snr_scaling(diameter_from, diameter_to, wavelength):
    """Determine the scaling constant `c` to convert the SNR generated by a
    `diameter_from` target to a `diameter_to` target. I.e. by
    `c SNR(diameter_from) = SNR(diameter_to)`.

    :param float diameter_from: object diameter of snr to convert from (meters)
    :param float diameter_to: object diameter of snr to convert to (meters)
    :param float wavelength: radar wavelength (meters)
    """

    sep_const = wavelength / (np.pi * np.sqrt(3.0))
    if diameter_from < sep_const and diameter_to < sep_const:
        scale = diameter_to**6 / diameter_from**6
    elif diameter_from < sep_const and diameter_to >= sep_const:
        scale = (wavelength**4 * diameter_to**2) / (9 * np.pi**4 * diameter_from**6)
    elif diameter_from >= sep_const and diameter_to < sep_const:
        scale = (9 * np.pi**4 * diameter_to**6) / (wavelength**4 * diameter_from**2)
    else:
        scale = diameter_to**6 / diameter_from**6

    return scale


def hard_target_rcs(wavelength, diameter):
    """Determine the radar cross section for a hard target.
    Assume a smooth transition between Rayleigh and optical scattering.
    Ignore Mie regime and use either optical or Rayleigh scatter.

    :param float wavelength: radar wavelength (meters)
    :param float/numpy.ndarray diameter: diameter in meters of the objects.
    """
    is_rayleigh = diameter < wavelength / (np.pi * np.sqrt(3.0))
    is_optical = diameter >= wavelength / (np.pi * np.sqrt(3.0))
    optical_rcs = np.pi * diameter**2.0 / 4.0
    rayleigh_rcs = diameter**6.0 * 9.0 * np.pi**5 / (4.0 * wavelength**4)
    rcs = is_rayleigh * rayleigh_rcs + is_optical * optical_rcs
    return rcs


@t.overload
def hard_target_snr(
    gain_tx: npt.NDArray[np.float64],
    gain_rx: npt.NDArray[np.float64],
    wavelength: float,
    power_tx: npt.NDArray[np.float64] | float,
    range_tx_m: npt.NDArray[np.float64],
    range_rx_m: npt.NDArray[np.float64],
    diameter=0.01,
    bandwidth: npt.NDArray[np.float64] | float = 10.0,
    rx_noise_temp: npt.NDArray[np.float64] | float = 150.0,
    radar_albedo=1.0,
) -> npt.NDArray[np.float64]: ...


@t.overload
def hard_target_snr(
    gain_tx: float,
    gain_rx: float,
    wavelength: float,
    power_tx: float,
    range_tx_m: float,
    range_rx_m: float,
    diameter=0.01,
    bandwidth=10.0,
    rx_noise_temp=150.0,
    radar_albedo=1.0,
) -> float: ...


def hard_target_snr(
    gain_tx: npt.NDArray[np.float64] | float,
    gain_rx: npt.NDArray[np.float64] | float,
    wavelength: float,
    power_tx: npt.NDArray[np.float64] | float,
    range_tx_m: npt.NDArray[np.float64] | float,
    range_rx_m: npt.NDArray[np.float64] | float,
    diameter=0.01,
    bandwidth: npt.NDArray[np.float64] | float = 10.0,
    rx_noise_temp: npt.NDArray[np.float64] | float = 150.0,
    radar_albedo=1.0,
):
    """
    Determine the signal-to-noise ratio (energy-to-noise) ratio for a hard target.
    Assume a smooth transition between Rayleigh and optical scattering.
    Ignore Mie regime and use either optical or Rayleigh scatter.

    :param float/numpy.ndarray gain_tx: transmit antenna gain, linear
    :param float/numpy.ndarray gain_rx: receiver antenna gain, linear
    :param float wavelength: radar wavelength (meters)
    :param float power_tx: transmit power (W)
    :param float/numpy.ndarray range_tx_m: range from transmitter to target (meters)
    :param float/numpy.ndarray range_rx_m: range from target to receiver (meters)
    :param float diameter: object diameter (meters)
    :param float bandwidth: effective receiver noise bandwidth
    :param float rx_noise_temp: receiver noise temperature (K)
    :return: signal-to-noise ratio
    :rtype: float/numpy.ndarray


    **Reference:** Markkanen et.al., 1999

    """

    is_rayleigh: bool = diameter < wavelength / (np.pi * np.sqrt(3.0))
    is_optical: bool = diameter >= wavelength / (np.pi * np.sqrt(3.0))
    rayleigh_power: npt.NDArray[np.float64] | float = (
        9.0
        * power_tx
        * (
            ((gain_tx * gain_rx) * (np.pi**2.0) * (diameter**6.0))
            / (256.0 * (wavelength**2.0) * (range_rx_m**2.0 * range_tx_m**2.0))
        )
    )
    optical_power: npt.NDArray[np.float64] | float = (
        power_tx
        * (((gain_tx * gain_rx) * (wavelength**2.0) * (diameter**2.0)))
        / (256.0 * (np.pi**2) * (range_rx_m**2.0 * range_tx_m**2.0))
    )
    rx_noise = scipy.constants.k * rx_noise_temp * bandwidth
    snr = ((is_rayleigh) * rayleigh_power + (is_optical) * optical_power) * radar_albedo / rx_noise
    return snr


def hard_target_diameter(
    gain_tx,
    gain_rx,
    wavelength,
    power_tx,
    range_tx_m,
    range_rx_m,
    snr,
    bandwidth=10,
    rx_noise_temp=150.0,
    radar_albedo=1.0,
):
    """
    Determine the diamtere for a hard target based on the signal-to-noise ratio (energy-to-noise).
    Assume a smooth transition between Rayleigh and optical scattering.
    Ignore Mie regime and use either optical or Rayleigh scatter.

    :param float/numpy.ndarray gain_tx: transmit antenna gain, linear
    :param float/numpy.ndarray gain_rx: receiver antenna gain, linear
    :param float wavelength: radar wavelength (meters)
    :param float power_tx: transmit power (W)
    :param float/numpy.ndarray range_tx_m: range from transmitter to target (meters)
    :param float/numpy.ndarray range_rx_m: range from target to receiver (meters)
    :param float snr: object signal to noise ratio (1)
    :param float bandwidth: effective receiver noise bandwidth
    :param float rx_noise_temp: receiver noise temperature (K)
    :return: diameter (meters)
    :rtype: float/numpy.ndarray


    **Reference:** Markkanen et.al., 1999

    """

    rx_noise = scipy.constants.k * rx_noise_temp * bandwidth
    power = snr * rx_noise / radar_albedo

    diameter_rayleigh = (
        256.0
        * (wavelength**2.0)
        * (range_rx_m**2.0 * range_tx_m**2.0)
        * power
        / (9.0 * power_tx * (gain_tx * gain_rx) * (np.pi**2.0))
    ) ** (1.0 / 6.0)
    diameter_optical = np.sqrt(
        256.0
        * (np.pi**2)
        * (range_rx_m**2.0 * range_tx_m**2.0)
        * power
        / (power_tx * (gain_tx * gain_rx) * (wavelength**2.0))
    )

    separatrix = wavelength / (np.pi * np.sqrt(3.0))

    is_rayleigh = np.logical_and(diameter_rayleigh < separatrix, diameter_optical < separatrix)
    is_optical = np.logical_and(diameter_rayleigh >= separatrix, diameter_optical >= separatrix)

    diameter = (is_rayleigh) * diameter_rayleigh + (is_optical) * diameter_optical

    return diameter


def incoherent_snr(p_s, p_n, epsilon=0.05, B=10.0, t_incoh=3600.0):
    """Calculate the incoherent SNR based on ????

    TODO: Finish docstring
    TODO: generalize theory??
    TODO: Juha knows
    """
    snr = p_s / p_n
    t_epsilon = ((p_s + p_n) ** 2.0) / (epsilon**2.0 * p_s**2.0 * B)

    K = t_incoh * B
    delta_pn = p_n / np.sqrt(K)
    snr_incoh = p_s / delta_pn

    return snr, snr_incoh, t_epsilon


def doppler_spread_hard_target_snr(
    t_obs,
    spin_period,
    gain_tx,
    gain_rx,
    wavelength,
    power_tx,
    range_tx_m,
    range_rx_m,
    duty_cycle=0.25,
    diameter=10.0,
    bandwidth=10,
    rx_noise_temp=150.0,
    radar_albedo=0.1,
):
    """
    t_obs = observation duration

    #TODO: Double check the "bandwidth" parameter to see that it is actually
    # defined and used correctly

    returns:
    snr - signal to noise ratio using coherent integration, when doing object discovery with a
          limited coherent integration duration and no incoherent integration
    snr_incoh - the signal to noise ratio using incoherent integration, when using a priori
                orbital elements to assist in coherent integration and incoherent integration.
                coherent integration length is determined by t_obs (seconds)
    """

    doppler_bandwidth = 4 * np.pi * diameter / (wavelength * spin_period)

    # for serendipitous discovery
    detection_bandwidth = np.max([doppler_bandwidth, bandwidth * duty_cycle, 1.0 / t_obs])

    # for detection with a periori know orbit
    # the bandwidth cannot be smaller than permitted by the observation duration.
    base_int_bandwidth = np.max([doppler_bandwidth, (1.0 / t_obs)])

    # standard one segment rx noise
    rx_noise = scipy.constants.k * rx_noise_temp * bandwidth

    # effective noise power when using just coherent integration
    p_n0 = scipy.constants.k * rx_noise_temp * detection_bandwidth / duty_cycle

    # effective noise power when doing incoherent integration and using a good
    # a priori orbital elements
    p_n1 = scipy.constants.k * rx_noise_temp * base_int_bandwidth / duty_cycle

    h_snr = hard_target_snr(
        gain_tx=gain_tx,
        gain_rx=gain_rx,
        wavelength=wavelength,
        power_tx=power_tx,
        range_tx_m=range_tx_m,
        range_rx_m=range_rx_m,
        diameter=diameter,
        bandwidth=bandwidth,
        rx_noise_temp=rx_noise_temp,
        radar_albedo=radar_albedo,
    )
    p_s = h_snr * rx_noise

    snr, snr_incoh, te = incoherent_snr(p_s, p_n1, B=base_int_bandwidth, t_incoh=t_obs)

    snr_coh = p_s / p_n0
    return snr_coh, snr_incoh



from . import (
    pointing_funcs,
    controller_base,
    fence_scan_controller,
    tracker_controller,
    sparse_tracker_controller,
)

from .controller_base import ControllerBase
from .fence_scan_controller import FenceScanController
from .tracker_controller import TrackerController
from .sparse_tracker_controller import SparseTrackerController



#!/usr/bin/env python

"""Define the concept of a radar scheduler that handles the RadarControler's and determines the behavior of the radar.

"""

from .scheduler import Scheduler
from .static_list import StaticList
from .tracking import Tracking, PriorityTracking

from .observed_parameters import ObservedParameters

from .pointing_schedule import PointingSchedule



#!/usr/bin/env python

"""

"""

import numpy as np

from .scheduler import Scheduler


class PointingSchedule(Scheduler):
    """#TODO: Docstring"""

    def generate_schedule(self, t, generator):
        rxp = []
        txp = []
        rx_pos = []
        tx_pos = []
        metas = []
        for ind, mrad in enumerate(generator):
            radar, meta = mrad

            metas.append(meta)

            rxp.append([])
            rx_pos.append([])
            for ri, rx in enumerate(radar.rx):
                rxp[-1].append(rx.pointing_ecef)
                rx_pos[-1].append(rx.ecef)

            txp.append([])
            tx_pos.append([])
            for ti, tx in enumerate(radar.tx):
                txp[-1].append(tx.pointing_ecef)
                tx_pos[-1].append(tx.ecef)

        data = {
            "t": t,
            "rx": rxp,
            "tx": txp,
            "rx_pos": rx_pos,
            "tx_pos": tx_pos,
            "meta": metas,
        }
        return data



#!/usr/bin/env python

"""Contains functions related to error calculation and a generalized method of generating random noisy data.

"""

from .errors import Errors

from .linearized_coded import LinearizedCodedIonospheric
from .linearized_coded import LinearizedCoded

from .ionospheric_ray_trace import calculate_delay, ray_trace, ray_trace_error, ionospheric_error
from .ionospheric_ray_trace import IonosphericRayTrace


from . import atmospheric_drag

from .linearized_orbit_determination import covariance_propagation, orbit_determination_covariance



#!/usr/bin/env python

"""Contains all helper functions to automate parallelization with MPI, handle caching and stepping of simulations."""

# Python standard import
import os
import pathlib
import shutil
from collections import OrderedDict
import logging
import copy
import pickle
import datetime
from glob import glob
import os
import argparse

# Third party import
import h5py
import numpy as np

try:
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
except ImportError:
    comm = None

logger = logging.getLogger(__name__)


def copy_tree(src, dst, linkfiles=False):
    """Copy path on thread rank=0 with :code:`shutil.copytree` or :code:`copy2` for files. If :code:`linkfiles` is true, files are soft-linked rather then copied."""
    if src.is_dir():
        if linkfiles:
            shutil.copytree(src, dst, copy_function=os.link)
        else:
            shutil.copytree(src, dst)
    else:
        if linkfiles:
            os.link(src, dst)
        else:
            shutil.copy2(src, dst)


def log_exceptions(func):
    """Enrich the exception error messages by logging the `args` and `kwargs`"""

    def wrapped_step(self, *args, **kwargs):
        try:
            rets = func(self, *args, **kwargs)
        except BaseException as err:
            logger.exception(f"\nargs: {args}\n kwargs: {kwargs}")
            raise err

        return rets

    return wrapped_step


def MPI_single_process(process_id):
    """Simulation step single process method restriction decorator.

    :param int process_id: The process id the wrapped function should only execute on. All other processes return :code:`None`.

    """

    def step_wrapping(func):
        def wrapped_step(self, *args, **kwargs):
            if comm is not None:
                if comm.rank == process_id:
                    rets = func(self, *args, **kwargs)
                else:
                    rets = None
            else:
                rets = func(self, *args, **kwargs)
            return rets

        if hasattr(func, "_simulation_step"):
            wrapped_step._simulation_step = func._simulation_step
        return wrapped_step

    return step_wrapping


def MPI_action(action, iterable=False, root=0):
    """Simulation step MPI post step action decorator.

    :param str action: Mode of operations on node-data communication, available options are "gather", "gather-clear", "bcast" and "barrier".
    :param bool iterable: Indicates if the "gather", "gather-clear" or "bcast" should consider an iterable (that has been parallelized with MPI).
    :param int root: The target MPI process for the "gather", "gather-clear" and for the source process for "bcast" if :code:`iterable=False`.
    """

    def step_wrapping(func):
        def wrapped_step(self, *args, **kwargs):
            rets = func(self, *args, **kwargs)

            if comm is None:
                return rets

            if iterable:
                mpi_inds = []
                for thrid in range(comm.size):
                    mpi_inds.append(range(thrid, len(rets), comm.size))

            if action == "barrier":
                comm.barrier()
            elif action.startswith("gather"):
                if iterable:
                    if comm.rank == root:
                        for thr_id in range(comm.size):
                            if thr_id != root:
                                for ind in mpi_inds[thr_id]:
                                    rets[ind] = comm.recv(source=thr_id, tag=ind)

                    else:
                        for ind in mpi_inds[comm.rank]:
                            comm.send(rets[ind], dest=root, tag=ind)

                    if action == "gather-clear":
                        if comm.rank != root:
                            for ind in mpi_inds[comm.rank]:
                                rets[ind] = None
                else:
                    all_rets = [None] * comm.size
                    all_rets[comm.rank] = rets

                    if comm.rank == root:
                        for thr_id in range(comm.size):
                            if thr_id != root:
                                all_rets[thr_id] = comm.recv(source=thr_id, tag=thr_id)
                    else:
                        comm.send(all_rets[comm.rank], dest=root, tag=comm.rank)

                    if action == "gather-clear":
                        if comm.rank != root:
                            all_rets[comm.rank] = None
                    rets = all_rets

            elif action == "bcast":
                if iterable:
                    for thr_id in range(comm.size):
                        for ind in mpi_inds[thr_id]:
                            rets[ind] = comm.bcast(rets[ind], root=thr_id)
                else:
                    rets = comm.bcast(rets, root=root)

            return rets

        if hasattr(func, "_simulation_step"):
            wrapped_step._simulation_step = func._simulation_step
        return wrapped_step

    return step_wrapping


def iterable_step(iterable, MPI=False, reduce=None):
    """Simulation step iteration decorator.

    :param str/list iterable: The name/list of names of the instance properties (fetched using :code:`getattr`) to iterate over. Can be multiple levels, e.g. :code:`object.subobject.a_list`.
    :param bool MPI: Determines if the iteration should be MPI-parallelized
    :param function reduce: A pointer to the binary-function used to reduce the results.
    """
    reduce_ = reduce

    if isinstance(iterable, str):
        iterable = [iterable]

    def step_wrapping(func):
        def wrapped_step(self, *args, **kwargs):
            all_attrs = []
            for var in iterable:
                subvars = var.split(".")
                for vari, subvar in enumerate(subvars):
                    if vari == 0:
                        obj = getattr(self, subvar)
                    else:
                        obj = getattr(obj, subvar)
                all_attrs.append(obj)

            if len(all_attrs) == 1:
                attr = all_attrs[0]
            else:
                attr = list(zip(*all_attrs))

            if MPI and comm is not None:
                _iter = list(range(comm.rank, len(attr), comm.size))
            else:
                _iter = list(range(len(attr)))

            if reduce_ is None:
                rets = [None] * len(attr)
            else:
                rets = None

            step_name = kwargs.get("_step_name", None)

            _iters = 0
            _total = len(_iter)

            for index in _iter:
                item = attr[index]

                if hasattr(func, "_simulation_step"):
                    kwargs["_iterable_index"] = index

                ret = func(self, index, item, *args, **kwargs)
                if reduce_ is None:
                    rets[index] = ret
                else:
                    rets = reduce_(rets, ret)
                    del ret

                _iters += 1

            return rets

        wrapped_step._simulation_step = True
        return wrapped_step

    return step_wrapping


def store_step(store, iterable=False):
    """Simulation step storing decorator

    :param str/list store: The name/list names of the properties to save the method return as (set using :code:`setattr`). Can be multiple levels, e.g. :code:`object.subobject.a_property`. The order of the names correspond to the order of method returned variables.
    :param bool iterable: Determines if the return of the method is an iteration or not. If its an iteration, it splits the return values into different lists based on the number of variables.
    """
    if isinstance(store, str):
        store = [store]

    def step_wrapping(func):
        def wrapped_step(self, *args, **kwargs):
            rets = func(self, *args, **kwargs)

            for si, var in enumerate(store):
                subvars = var.split(".")
                if len(subvars) == 1:
                    obj = self
                    name = subvars[0]
                else:
                    for vari, subvar in enumerate(subvars[:-1]):
                        if vari == 0:
                            obj = getattr(self, subvar)
                        else:
                            obj = getattr(obj, subvar)
                    name = subvars[-1]

                if iterable:
                    iter_obj = [None] * len(rets)
                    for index in range(len(rets)):
                        if len(store) == 1 or rets[index] is None:
                            iter_obj[index] = rets[index]
                        else:
                            iter_obj[index] = rets[index][si]
                    setattr(obj, name, iter_obj)
                else:
                    if len(store) == 1 or rets is None:
                        setattr(obj, name, rets)
                    else:
                        setattr(obj, name, rets[si])

            return rets

        wrapped_step._simulation_step = True
        return wrapped_step

    return step_wrapping


def iterable_cache(steps, caches, MPI=False, reduce=None):
    """Simulation step cache iteration decorator

    :param str/list steps: The name/list of names of the cached steps to be iterated over. It uses the step name to find the files in the corresponding folder.
    :param str/list caches: The name/list of cache-methods to be used to load the caches of the steps.
    :param bool MPI: Determines if the iteration should be MPI-parallelized.
    :param function reduce: A pointer to the binary-function used to reduce the results.
    """
    reduce_ = reduce

    if isinstance(steps, str):
        steps = [steps]
    if isinstance(caches, str):
        caches = [caches]

    def step_wrapping(func):
        def wrapped_step(self, *args, **kwargs):
            all_files = []
            all_indecies = []
            for step, cache in zip(steps, caches):
                dir_ = self.get_path(step)
                if not dir_.is_dir():
                    raise ValueError(f"Input step {step} has no cache {cache}")
                files = [pathlib.Path(x) for x in glob(str(dir_ / f"*"))]
                indecies = [int(file.stem.split("_")[0]) for file in files]

                files = [x for _, x in sorted(zip(indecies, files), key=lambda pair: pair[0])]
                indecies = sorted(indecies)

                all_files += [files]
                all_indecies += [indecies]

            if len(steps) == 1:
                files_lst = [[x] for x in all_files[0]]
            else:
                # check all inds exist
                for index, inds in enumerate(zip(*all_indecies)):
                    if len(set(inds)) > 1:
                        raise ValueError(f"Missing index {inds}")
                files_lst = list(zip(*all_files))
            indecies = all_indecies[0]

            if MPI and comm is not None:
                _iter = list(range(comm.rank, len(indecies), comm.size))
            else:
                _iter = list(range(len(indecies)))

            if reduce_ is None:
                rets = [None] * len(indecies)
            else:
                rets = None

            step_name = kwargs.get("_step_name", None)

            _iters = 0
            _total = len(_iter)

            for index__ in _iter:
                index = indecies[index__]
                files = files_lst[index__]

                item = []
                for cache, fname in zip(caches, files):
                    lfunc = getattr(self, f"load_{cache}")
                    item += [lfunc(fname)]

                if len(caches) == 1:
                    item = item[0]

                if hasattr(func, "_simulation_step"):
                    kwargs["_iterable_index"] = index

                ret = func(self, index, item, *args, **kwargs)
                if reduce_ is None:
                    rets[index__] = ret
                else:
                    rets = reduce_(rets, ret)
                    del ret

                _iters += 1

            return rets

        wrapped_step._simulation_step = True
        return wrapped_step

    return step_wrapping


def pre_post_actions(pre=None, post=None):
    """Executes the methods given by name pre- and post-execution of the wrapped function. The methods all get the same input arguments as the wrapped function and their return is discarded.

    :param str pre: A list of strings or a single string for the name(s) of the class method(s) to be executed before the wrapped method.
    :param str post: A list of strings or a single string for the name(s) of the class method(s) to be executed after the wrapped method.
    """

    if isinstance(pre, str):
        pre = [pre]
    if isinstance(post, str):
        post = [post]

    def step_wrapping(func):
        def wrapped_step(self, *args, **kwargs):
            if pre is not None:
                for func_name in pre:
                    pre_func = getattr(self, func_name)
                    pre_func(*args, **kwargs)

            ret = func(self, *args, **kwargs)

            if post is not None:
                for func_name in post:
                    post_func = getattr(self, func_name)
                    post_func(*args, **kwargs)

            return ret

        if hasattr(func, "_simulation_step"):
            wrapped_step._simulation_step = func._simulation_step
        return wrapped_step

    return step_wrapping


def cached_step(caches):
    """Simulation step caching decorator


    :param str caches: Is a list of strings for the caches to be used, available by default is "h5" and "pickle". Custom caches are implemented by implementing methods with the string name but prefixed with :code:`load_` and :code:`save_`.

    """
    if isinstance(caches, str):
        caches = [caches]

    def step_wrapping(func):
        def wrapped_step(self, *args, **kwargs):
            step = kwargs.get("_step_name", "cached_data")
            fname_parts = kwargs.pop("_fname_parts", ["data"])
            index = kwargs.get("_iterable_index", None)
            if index is None:
                index_lst = []
            else:
                index_lst = [str(index)]

            dir_ = self.get_path(step)

            loaded_ = False

            # load
            for cache in caches:
                fname = dir_ / f'{"_".join(index_lst + fname_parts)}.{cache}'
                if fname.is_file():
                    lfunc = getattr(self, f"load_{cache}")
                    try:
                        ret = lfunc(fname)
                        loaded_ = True
                    except (OSError, EOFError, UnicodeError):
                        fname.unlink()

                if loaded_:
                    break

            # if there are no caches
            if not loaded_:
                ret = func(self, *args, **kwargs)

                # save
                for cache in caches:
                    fname = dir_ / f'{"_".join(index_lst + fname_parts)}.{cache}'
                    sfunc = getattr(self, f"save_{cache}")
                    sfunc(fname, ret)

            return ret

        wrapped_step._simulation_step = True
        return wrapped_step

    return step_wrapping


# TODO: refactor
#  notes from danielk:
#    I kind of like the idea of this class from the sense that it aims to make
#    simulation building much easier since there are always pattern you will need
#    like caching output of steps and trivially parallelizing iterations
#    but it will need to be refactored now that we change the rest - i think this is one of the functions
#    that might get cut once we do that
#
#   i think this whole scheduler part will be refactored away anyway
#   now that i look a bit at this class i think we should instead split a lot of this out into
#   "lego" pieces that can be used and not force the structure of the class itself anyway
class Simulation:
    """Convenience simulation handler, creates a step-by-step simulation sequence and creates file system structure for saving of data to disk.

    :param Scheduler scheduler: A scheduler instance to run.
    :param str/pathlib.Path root: The path to the root folder where all files will be stored.
    """

    def __init__(self, scheduler, root, **kwargs):
        self.steps = OrderedDict()
        self.scheduler = scheduler

        if root is None:
            self.persistancy = False
        else:
            self.persistancy = True

        self.root = root

        self.__user_args = []

        if self.persistancy:
            if not isinstance(self.root, pathlib.Path):
                self.root = pathlib.Path(self.root)

            if comm is not None:
                if comm.rank == 0:
                    if not self.root.is_dir():
                        self.root.mkdir(exist_ok=True)
                comm.barrier()
            else:
                if not self.root.is_dir():
                    self.root.mkdir(exist_ok=True)

        self.branch_name = "master"

        if self.persistancy:
            _master = self.root / self.branch_name

            if comm is not None:
                if comm.rank == 0:
                    if not _master.is_dir():
                        _master.mkdir(exist_ok=True)
                comm.barrier()
            else:
                if not _master.is_dir():
                    _master.mkdir(exist_ok=True)

            self.make_paths()

            if comm is not None:
                comm.barrier()

        if self.scheduler is not None:
            self.scheduler.logger = logger

    def save_pickle(self, path, data):
        with open(path, "wb") as h:
            pickle.dump(data, h)

    def load_pickle(self, path):
        if path.is_file():
            with open(path, "rb") as h:
                ret = pickle.load(h)
            return ret

    def save_npy(self, path, data):
        np.save(path, data)

    def load_npy(self, path):
        if path.is_file():
            return np.load(path)

    def save_npz(self, path, data):
        np.savez(path, **data)

    def load_npz(self, path):
        if path.is_file():
            return np.load(path)

    def save_h5(self, path, data):
        with h5py.File(path, "w") as h:
            if isinstance(data, dict):
                for key in data:
                    if isinstance(data[key], np.ndarray):
                        h.create_dataset(key, data=data[key])
                    else:
                        h.attrs[key] = data[key]
            if isinstance(data, list) or isinstance(data, tuple):
                for ind in range(len(data)):
                    h.attrs["__ret_len"] = len(data)
                    if isinstance(data[ind], np.ndarray):
                        h.create_dataset(f"saved_data__{ind}", data=data[ind])
                    else:
                        h.attrs[f"saved_data__{ind}"] = data[ind]
            else:
                h.create_dataset("__saved_data", data=data)

    def load_h5(self, path):
        if path.is_file():
            with h5py.File(path, "r") as h:
                if "__saved_data" in h:
                    ret = h["__saved_data"][()].copy()

                elif "__ret_len" in h.attrs:
                    ret = [None] * copy.copy(h.attrs["__ret_len"])
                    for ind in range(len(ret)):
                        key = f"saved_data__{ind}"
                        if key in h:
                            ret[ind] = h[key][()].copy()
                        else:
                            ret[ind] = copy.copy(h.attrs[key])

                else:
                    ret = {}
                    for key in h:
                        ret[key] = h[key][()].copy()
                    for key in h.attrs:
                        ret[key] = copy.copy(h.attrs[key])
            return ret

    def make_paths(self):
        """Make all the folder for the current branch according to :code:`self.paths`."""
        for path in self.paths:
            abs_path = self.get_path(path)

            if comm is not None:
                if comm.rank == 0:
                    if not abs_path.is_dir():
                        abs_path.mkdir(exist_ok=True)
                comm.barrier()
            else:
                if not abs_path.is_dir():
                    abs_path.mkdir(exist_ok=True)

    @property
    def paths(self):
        """List of the name of all folders"""
        return [key for key in self.steps] + ["logs"]

    @property
    def log_path(self):
        """Path to the current log-output folder"""
        if self.persistancy:
            return self.root / self.branch_name / "logs"
        else:
            return None

    def get_path(self, name=None):
        """Given a relative file path, get the absolute path including root and branch."""
        if name is None:
            return self.root / self.branch_name
        else:
            return self.root / self.branch_name / name

    def delete(self, branch):
        """Delete branch."""

        if comm is not None:
            if comm.rank == 0:
                if (self.root / branch).is_dir():
                    shutil.rmtree(self.root / branch)
            comm.barrier()
        else:
            if (self.root / branch).is_dir():
                shutil.rmtree(self.root / branch)

        if self.branch_name == branch:
            if comm is not None:
                if comm.rank == 0:
                    (self.root / self.branch_name).mkdir(exist_ok=True)
                comm.barrier()
            else:
                (self.root / self.branch_name).mkdir(exist_ok=True)

            self.make_paths()

    @MPI_action(action="barrier")
    @MPI_single_process(process_id=0)
    def _make_branch(self, name, empty=False, linkfiles=None):
        """Create branch by creating a copy of the current branch state and checkout that branch. If the branch exists, just checkout that branch.

        :param str name: Name of the new branch
        :param bool empty: If :code:`True` do not copy the state of the current branch.
        :param list/bool linkfiles: If a list of paths that should be soft-linked rather then copied. If :code:`True`, soft-link all files.
        :return: None
        """
        if not self.persistancy:
            raise ValueError("Cannot use branches if no root is given")

        if (self.root / name).is_dir():
            logger.info(f'Branch "{name}" already exists')
        else:
            if empty:
                (self.root / name).mkdir(exist_ok=True)
                for path in self.paths:
                    (self.root / name / path).mkdir(exist_ok=True)

            else:
                if linkfiles is None:
                    copy_tree(self.root / self.branch_name, self.root / name, linkfiles=False)
                elif isinstance(linkfiles, list):
                    listing = pathlib.Path(self.root / self.branch_name).glob("./*")
                    for pth in listing:
                        if pth.name in linkfiles:
                            copy_tree(pth, self.root / name / pth.name, linkfiles=True)
                        else:
                            copy_tree(pth, self.root / name / pth.name, linkfiles=False)
                elif linkfiles:
                    copy_tree(self.root / self.branch_name, self.root / name, linkfiles=True)
                else:
                    raise TypeError(f'linkfiles type "{type(linkfiles)}" not supported')

    def branch(self, name, empty=False, linkfiles=None):
        """Create branch by creating a copy of the current branch state and checkout that branch. If the branch exists, just checkout that branch.

        :param str name: Name of the new branch
        :param bool empty: If :code:`True` do not copy the state of the current branch.
        :param list/bool linkfiles: If a list of paths that should be soft-linked rather then copied. If :code:`True`, soft-link all files.
        :return: None
        """
        self._make_branch(name, empty=empty, linkfiles=linkfiles)
        self.checkout(name)

    def status(self):
        """Prints the status"""
        raise NotImplementedError(
            "should contain branch-list, branch sizes, step lists and logs avalible"
        )

    def checkout(self, branch):
        """Change to branch."""
        self.branch_name = branch
        if self.persistancy:
            if not self.log_path.is_dir():
                self.log_path.mkdir(exist_ok=True)

    @log_exceptions
    def run(self, step=None, *args, **kwargs):
        """Run specific step with the supplied arguments or run all steps"""

        if self.persistancy:
            self.make_paths()

        if step is None:
            for name, func in self.steps.items():
                logger.info(f"Simulation:run:{name}")

                if hasattr(func, "_simulation_step"):
                    func(*args, _step_name=name, **kwargs)
                else:
                    func(*args, **kwargs)

                logger.info(f"Simulation:run:{name} [completed]")
        else:
            func = self.steps[step]

            logger.info(f"Simulation:run:{step}")
            if hasattr(func, "_simulation_step"):
                func(*args, _step_name=step, **kwargs)
            else:
                func(*args, **kwargs)

            logger.info(f"Simulation:run:{step} [completed]")

    def add_cmd_argument(self, *args, **kwargs):
        """Takes same arguments and keyword arguments as `parser.add_argument` but does not add the argument to the parser until `parse_cmd` is called.

        All arguments added here ends up as keyword arguments to the steps, the argument name used is always .
        """
        self.__user_args.append((args, kwargs))

    def parse_cmd(self):
        """Parses the arguments from a terminal command-line execution of the simulation and executes appropriately"""

        logger.info("Simulation:parse_cmd:parsing command")

        arg_parser = argparse.ArgumentParser(description="Simulation command-line interface")

        subparsers = arg_parser.add_subparsers(dest="sim_action", help="Actions")

        run_parser = subparsers.add_parser("run", help="Run simulation")
        cmd_parser = subparsers.add_parser("cmd", help="Simulation command")
        cmd_subparsers = cmd_parser.add_subparsers(dest="sim_cmd", help="Commands")

        if self.persistancy:
            run_parser.add_argument(
                "branch",
                type=str,
                help="Name of the branch to operate on",
            )
            branch_parser = cmd_subparsers.add_parser("branch", help="Branch simulation")
            branch_parser.add_argument(
                "name",
                type=str,
                help="Name of new branch",
            )
            branch_parser.add_argument(
                "--source",
                type=str,
                default="",
                help="Name of branch to clone onto new branch",
            )
            branch_parser.add_argument(
                "--link-branch",
                action="store_true",
                help="Soft-link all current files towards [source] branch",
            )
            delete_parser = cmd_subparsers.add_parser("delete", help="Delete branch")
            delete_parser.add_argument(
                "name",
                type=str,
                help="Branch name to delete",
            )

        run_parser.add_argument(
            "step",
            type=str,
            help='Name of the step to execute, if "all" executes all steps.',
        )

        dests = []
        for args, kwargs in self.__user_args:
            run_parser.add_argument(*args, **kwargs)
            action = run_parser._actions[-1]
            if hasattr(action, "dest"):
                dests.append(action.dest)

        args = arg_parser.parse_args()

        logger.debug(f"Simulation:parse_cmd:args = {args}")

        if args.sim_action == "cmd":
            if args.sim_cmd == "branch":
                if len(args.source) > 0:
                    self.checkout(args.source)
                    _empty = False
                else:
                    _empty = True

                if args.link_branch:
                    linkfiles = True
                else:
                    linkfiles = None

                self.branch(name=args.name, empty=_empty, linkfiles=linkfiles)
            elif args.sim_cmd == "delete":
                self.delete(args.name)

        elif args.sim_action == "run":
            if args.step == "all":
                step = None
            else:
                step = args.step

            if self.persistancy:
                self.checkout(args.branch)

            run_kwargs = dict()
            for dest in dests:
                run_kwargs[dest] = getattr(args, dest)

            self.run(step=step, **run_kwargs)



#!/usr/bin/env python

"""
   Monte-Carlo sampling of errors due to atmospheric drag force uncertainty.

   Estimate a power-law model of error standard deviation in along-track direction (largest error).

   Juha Vierinen
"""
import time
import numpy as n
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D


from ..space_object import SpaceObject as so


def get_inertial_basis(ecef0, ecef0_dt):
    """
    Given pos vector, and pos vector at a small positive time offset,
    calculate unit vectors for along track, normal (towards center of Earth), and cross-track directions
    """
    along_track = ecef0_dt - ecef0
    along_track = along_track / n.sqrt(
        along_track[0, :] ** 2.0 + along_track[1, :] ** 2.0 + along_track[2, :] ** 2.0
    )
    normal = ecef0 / n.sqrt(ecef0[0, :] ** 2.0 + ecef0[1, :] ** 2.0 + ecef0[2, :] ** 2.0)
    cross_track = n.copy(normal)
    cross_track[:, :] = 0.0
    cross_track[0, :] = along_track[1, :] * normal[2, :] - along_track[2, :] * normal[1, :]
    cross_track[1, :] = along_track[2, :] * normal[0, :] - along_track[0, :] * normal[2, :]
    cross_track[2, :] = along_track[0, :] * normal[1, :] - along_track[1, :] * normal[0, :]
    cross_track = cross_track / n.sqrt(
        cross_track[0, :] ** 2.0 + cross_track[1, :] ** 2.0 + cross_track[2, :] ** 2.0
    )
    return (along_track, normal, cross_track)


def atmospheric_errors(o, a_err_std=0.01, N_samps=100, plot=False, threshold_error=100.0, res=500):
    """
    Estimate position errors as a function of time, assuming
    a certain error in atmospheric drag.
    """

    t = 10 ** (n.linspace(2, 6.2, num=100))
    t_dt = n.copy(t) + 1.0
    ecef = o.get_state(t)

    print("n_days %d" % (n.max(t) / 24.0 / 3600.0))
    C_D0 = o.parameters["C_D"]
    err = n.copy(ecef)
    err[:, :] = 0.0

    t0 = time.time()

    for i in range(N_samps):
        o1 = o.copy()
        o1.mu0 = n.random.rand(1) * 360.0

        ecef = o1.get_state(t)
        ecef_dt = o1.get_state(t_dt)
        at, norm, ct = get_inertial_basis(ecef, ecef_dt)

        C_D = C_D0 + C_D0 * n.random.randn(1)[0] * a_err_std
        o1.parameters["C_D"] = C_D

        ecef1 = o1.get_state(t)

        err_now = ecef1 - ecef
        err[0, :] += (
            n.abs(err_now[0, :] * at[0, :] + err_now[1, :] * at[1, :] + err_now[2, :] * at[2, :])
            ** 2.0
        )

        # difference in radius is the best estimate for radial distance error.
        err[1, :] += n.abs(
            n.sqrt(ecef[0, :] ** 2.0 + ecef[1, :] ** 2.0 + ecef[2, :] ** 2.0)
            - n.sqrt(ecef1[0, :] ** 2.0 + ecef1[1, :] ** 2.0 + ecef1[2, :] ** 2.0)
        )
        #       and not this:  err[1,:]+=n.abs(err_now[0,:]*norm[0,:]+err_now[1,:]*norm[1,:]+err_now[2,:]*norm[2,:])**2.0
        err[2, :] += (
            n.abs(err_now[0, :] * ct[0, :] + err_now[1, :] * ct[1, :] + err_now[2, :] * ct[2, :])
            ** 2.0
        )

        elps = time.time() - t0
        print(
            "{}/{} done - time elapsed {:<5.2f} h | estimated time remaining {:<5.2f}".format(
                i + 1,
                N_samps,
                elps / 3600.0,
                elps / float(i + 1) * float(N_samps - i - 1) / 3600.0,
            )
        )

    ate = n.sqrt(err[0, :] / N_samps)
    if n.max(ate) > threshold_error:
        idx0 = n.where(ate > threshold_error)[0][0]
        days = t / 24.0 / 3600.0
        hour0 = 24.0 * days[idx0]
    else:
        hour0 = n.max(t / 3600.0)

    alpha = (n.log(err[0, -1] / N_samps) - n.log(err[0, 46] / N_samps)) / (
        n.log(t[-1]) - n.log(t[46])
    )
    # (n.log(t[-1])-n.log(t[0]))*alpha=n.log(err[0,-1]/N_samps)

    offset = n.log(err[0, 46] / N_samps)
    t1 = t[46]
    var = n.exp((n.log(t) - n.log(t[46])) * alpha + offset)

    if plot:
        plt.loglog(t / 24.0 / 3600.0, n.sqrt(err[0, :] / N_samps), label="Along track")
        plt.loglog(t / 24.0 / 3600.0, n.sqrt(var), label="Fit", alpha=0.5)
        plt.loglog(t / 24.0 / 3600.0, n.sqrt(err[1, :] / N_samps), label="Radial")
        plt.loglog(t / 24.0 / 3600.0, n.sqrt(err[2, :] / N_samps), label="Cross-track")
        if n.max(ate) > threshold_error:
            plt.axvline(days[idx0])
            plt.text(days[idx0], threshold_error, "$\\tau=%1.1f$ hours" % (24 * days[idx0]))
        plt.grid()
        plt.axvline(n.max(t) / 24.0 / 3600.0)
        plt.xlim([0, n.max(t) / 24.0 / 3600.0])
        plt.legend()
        plt.ylabel("Cartesian position error (m)")
        plt.xlabel("Time (days)")
        # plt.title("Atmospheric drag uncertainty related errors"%(alpha))
        plt.title(
            "a %1.0f (km) e %1.2f i %1.0f (deg) aop %1.0f (deg) raan %1.0f (deg)\nA %1.2f$\pm$ %d%% (m$^2$) mass %1.2f (kg)\n$\\alpha=%1.1f$ $t_1=%1.1f$ $\\beta=%1.1f$"
            % (
                o.orbit.a,
                o.orbit.e,
                o.orbit.i,
                o.orbit.omega,
                o.orbit.Omega,
                o.parameters["A"],
                int(a_err_std * 100.0),
                o.parameters["m"],
                alpha,
                t1,
                offset,
            )
        )

    return (hour0, offset, t1, alpha)



import numpy as np


max_datetime64_us = np.datetime64(np.iinfo(np.int64).max, "us")
min_datetime64_us = np.datetime64(np.iinfo(np.int64).min + 1, "us")  # +1 is needed, otherwise it will be NaT; fmt: skip;



import numpy as np
import xarray as xr
from sorts import scheduling
from sorts.scheduling import Schedule
from sorts.scheduling.priority_scheduling import priority_scheduling


def setup_function():
    # a workaround to avoid pytest printings and test printings interleaved in the same line
    # https://github.com/pytest-dev/pytest/issues/8574#issuecomment-1806404215
    print()

    import pandas as pd

    pd.set_option("display.expand_frame_repr", False)


def priority_scheduling_interleaved_schedule_test():
    _SK = scheduling._K

    # 30min long, 2hr intv
    sch_a = scheduling.from_ndarrays(
        start_time=np.arange(
            np.datetime64("2025-01-01", "s"),
            np.datetime64("2025-01-02", "s"),
            np.timedelta64(3600 * 2, "s"),
        ),
        end_time=np.arange(
            np.datetime64("2025-01-01 00:30:00", "s"),
            np.datetime64("2025-01-02 00:00:01", "s"),
            np.timedelta64(3600 * 2, "s"),
        ),
        exp_num=np.full(12, 0, dtype=np.int16),
        stn_num=np.full(12, 0, dtype=np.int16),
        simult_num=np.full(12, 0, dtype=np.int16),
        pointing=np.full((3, 12), 0.0, dtype=np.float64),
    )
    sch_a = Schedule(sch_a)

    # 1hr long, 1hr intv
    sch_b = scheduling.from_ndarrays(
        start_time=np.arange(
            np.datetime64("2025-01-01", "s"),
            np.datetime64("2025-01-02", "s"),
            np.timedelta64(3600, "s"),
        ),
        end_time=np.arange(
            np.datetime64("2025-01-01 01:00:00", "s"),
            np.datetime64("2025-01-02 00:00:01", "s"),
            np.timedelta64(3600, "s"),
        ),
        exp_num=np.full(24, 1, dtype=np.int16),
        stn_num=np.full(24, 0, dtype=np.int16),
        simult_num=np.full(24, 0, dtype=np.int16),
        pointing=np.full((3, 24), 1.0, dtype=np.float64),
    )
    sch_b = Schedule(sch_b)

    resultant_sch = priority_scheduling([sch_a, sch_b], {0: [(0, 0)], 1: [(0, 0)]})

    assert all(xr.ufuncs.equal(resultant_sch[_SK.exp_num], [0, 1] * 12))

    return



"""
Shared types in this subpackage.

(Types might live in their own module instead of here if it improves readability,
and the imports can be worked around, e.g, by `if t.TYPE_CHECKING`)

# todo: probably move this out to the other types files??
"""

from dataclasses import dataclass
from sorts import types, radar
from sorts.space_object import SpaceObject


# TODO: can be combined with type 'Passage'?
@dataclass(kw_only=True)
class Passage:
    """Represent a passage of a space object over the field of view of a single TX, multiple simultaneous RX radar stations."""

    space_object: SpaceObject
    tx_station: radar.Station
    rx_stations: list[radar.Station]
    epoch: types.Datetime64_us
    time_range: types.TimeRange_us
    """The start time and end time of the passage, a right-open interval"""


SpaceObjectJacobianTuple = tuple[
    SpaceObject, SpaceObject, SpaceObject, SpaceObject, SpaceObject, SpaceObject, SpaceObject
]
"""A tuple of 7 `SpaceObject`, the first one is the original one and the next 6 are perturbated versions."""



import numpy.typing as npt
import pandas as pd
import bokeh.models as bokeh_models
import bokeh.layouts as bokeh_layouts
from sorts import scheduling
from sorts.types import EcefStates, Datetime64_us
from sorts.scheduling import Schedule
from .schedule_plot import _schedule_plot_from_cds
from .azel_skyplot import _azel_skyplot_cds_cols, _azel_skyplot_from_cds
from .ecef_states_positions_plot import (
    _ecef_states_positions_plot_cds_cols,
    _ecef_states_positions_plot_from_cds,
)


def _radar_schedule_ecef_position_plot_cds_df(
    ecefs: EcefStates, ecefs_time: npt.NDArray[Datetime64_us], sch: Schedule
):
    _SK = scheduling._K

    df = scheduling.to_dataframe(sch)

    # bokeh requires str type for categorical axis
    df[_SK.exp_num] = df[_SK.exp_num].astype(str)

    # insert columns for azel_skyplot
    azel_skyplot_cols = _azel_skyplot_cds_cols(
        sch[_SK.pointing][0].to_numpy(), sch[_SK.pointing][1].to_numpy()
    )
    for k, v in azel_skyplot_cols.items():
        df[k] = v

    # merging schedule and ecefs into the same dataframe
    ecef_pos_plot_cols = _ecef_states_positions_plot_cds_cols(ecefs)
    df = pd.merge(
        df,
        pd.DataFrame({_SK.start_time: ecefs_time, **ecef_pos_plot_cols}),
        on=_SK.start_time,
        how="outer",
    )

    return df


def radar_schedule_ecef_position_plot(
    ecefs: EcefStates, ecefs_time: npt.NDArray[Datetime64_us], sch: Schedule
):
    _SK = scheduling._K

    df = _radar_schedule_ecef_position_plot_cds_df(ecefs=ecefs, ecefs_time=ecefs_time, sch=sch)

    cds = bokeh_models.ColumnDataSource(df)

    start_time = df[_SK.start_time].min()
    end_time = df[_SK.end_time].max()

    sch_plot, sch_plot_bar, *_ = _schedule_plot_from_cds(
        cds,
        start_time=start_time,
        end_time=end_time,
        y_range=df[_SK.exp_num].dropna().unique(),
    )
    sch_plot_bar_select_tool = bokeh_models.BoxSelectTool()
    sch_plot_bar.add_tools(sch_plot_bar_select_tool)
    sch_plot_bar.toolbar.active_drag = sch_plot_bar_select_tool

    skyplot = _azel_skyplot_from_cds(cds)
    skyplot_select_tool = bokeh_models.LassoSelectTool()
    skyplot.add_tools(skyplot_select_tool)
    skyplot.toolbar.active_drag = skyplot_select_tool

    ecefpos_plot = _ecef_states_positions_plot_from_cds(cds)
    ecefpos_plot_select_tool = bokeh_models.LassoSelectTool()
    ecefpos_plot.add_tools(ecefpos_plot_select_tool)

    plot = bokeh_layouts.layout(
        [
            [sch_plot],
            [skyplot, ecefpos_plot],
        ]  # type: ignore
    )

    return plot



import numpy as np
import numpy.typing as npt
from sorts.space_object import SpaceObject
from sorts.types import Datetime_Like, Datetime64_us, Timedelta64_us, Float64_as_sec
from sorts.utils import to_datetime64_us
from .ecef_states_positions_plot import ecef_states_positions_plot


def kepler_space_object_on_map(
    space_object: SpaceObject,
    epoch: Datetime_Like,
    num_points=500,
    start_time: Datetime_Like | None = None,
    end_time: Datetime_Like | None = None,
):
    """Plot a space object with keplerian orbit on a map in mercator projection"""

    # TODO: check if space_object.orbit.period can be adj to always return a float
    _period = space_object.orbit.period
    period: np.timedelta64
    match _period:
        case np.ndarray():
            period = _period[0].astype("timedelta64[s]")
        case int() | float():
            period = np.timedelta64(int(_period), "s")
        case _:
            raise RuntimeError(
                f"`space_object.orbit.period` have to be of type float or ndarray of float64, but is in type {type(space_object.orbit.period)}"
            )

    period = np.timedelta64(period, "s")
    epoch = to_datetime64_us(epoch)
    start_time = to_datetime64_us(start_time) if start_time is not None else epoch
    end_time = to_datetime64_us(end_time) if end_time is not None else start_time + period

    time_arr: npt.NDArray[Datetime64_us] = np.linspace(
        start_time.astype(np.float64),
        end_time.astype(np.float64),
        num_points,
    ).astype("datetime64[us]")

    dt_arr: npt.NDArray[Timedelta64_us] = time_arr - to_datetime64_us(epoch)
    dsec_arr: npt.NDArray[Float64_as_sec] = dt_arr.astype(np.float64) / 1e6  # type: ignore

    ecefs = space_object.get_state(dsec_arr)
    plot = ecef_states_positions_plot(ecefs)

    return plot



from sorts.population import Population
import bokeh.models as bokeh_models


# TODO: re-eval if we need this
def space_object_population_table_plot(population: Population):
    source = bokeh_models.ColumnDataSource(data={f: population[f] for f in population.fields})
    columns = [bokeh_models.TableColumn(field=f, title=f) for f in population.fields]
    data_table = bokeh_models.DataTable(
        source=source,
        columns=columns,
        width=800,
        height=300,
        # selectable=True,
        # selectable="checkbox",
    )

    return data_table



#!/usr/bin/env python

"""Defines the concept of a survey or "scanning" pattern as a function of time for a radar station.
Has a number of predefined subclasses such as "fence" or "beampark".

"""

from .scan import Scan

from .fence import Fence
from .random_uniform import RandomUniform
from .uniform import Uniform
from .plane import Plane
from .bp import Beampark

__all__ = [
    "Scan",
    "Plane",
    "Uniform",
    "RandomUniform",
    "Fence",
    "Beampark",
]



from __future__ import annotations
import logging, typing as t, pickle
from pathlib import Path
import numpy as np
import numpy.typing as npt
import xarray as xr
import pyorb
import sorts
from tqdm import tqdm
from sorts import scheduling, controller, simulation
from sorts.types import Datetime_Like, Float64_as_sec, Datetime64_us, EcefStates
from sorts.utils import to_datetime64_us
from sorts.radar import Station, StationId
from sorts.simulation import Passage
from sorts.simulation.funcs import InterpolatedPropagation
from sorts.scheduling import Schedule, ExperimentDetailMap
from sorts.simulation.stx_mrx_simulation.simulation_unit import (
    SimulationUnit,
    FromPassagesOverTxRxStationPairParam,
    Observation,
)

logger = logging.getLogger(__name__)


sim_unit_fname_tpl = "sim_unit.{id}.pickle"


class SpaceObjectDsecSampler(t.Protocol):
    def __call__(
        self,
        orbit: pyorb.Orbit,
        epoch: Datetime_Like,
        start_time: Datetime_Like,
        end_time: Datetime_Like,
    ) -> npt.NDArray[Float64_as_sec]: ...


def sample_and_propagate_space_objects_states(
    sampler: SpaceObjectDsecSampler,
    spobjs: t.Sequence[sorts.SpaceObject],
    start_time: Datetime64_us,
    end_time: Datetime64_us,
) -> tuple[list[npt.NDArray[Float64_as_sec]], list[EcefStates]]:
    """
    Use the sampler to get the delta time of space object within the simulation `start_time` and `end_time`,
    then get the space object states at those delta time using the propagator in the space object.

    Returns a list of sampled delta seconds and a list of corresponding states.
    """

    spobjs_smpl_dsec: list[npt.NDArray[Float64_as_sec]] = []
    for spobj in tqdm(spobjs, desc="sampling spobjs dt", total=len(spobjs)):
        spobjs_smpl_dsec.append(
            sampler(spobj.state, to_datetime64_us(spobj.epoch), start_time, end_time)
        )

    spobjs_smpl_states: list[EcefStates] = []
    for spobj, spobj_smpl_dsec in tqdm(
        zip(spobjs, spobjs_smpl_dsec),
        desc="propagating spobjs states at sampled dt",
        total=len(spobjs),
    ):
        spobjs_smpl_states.append(spobj.get_state(spobj_smpl_dsec))

    return spobjs_smpl_dsec, spobjs_smpl_states


def group_passages_by_tx_rx_station_pair(
    passages: t.Sequence[Passage],
) -> dict[tuple[StationId, StationId], list[Passage]]:
    groupped_passages: dict[tuple[StationId, StationId], list[Passage]] = {}

    for passage in passages:
        # todo: make sure this is not broken
        tx_station_id = passage.tx_station.uid
        rx_station_id = passage.rx_stations[0].uid

        if (tx_station_id, rx_station_id) in groupped_passages:
            groupped_passages[(tx_station_id, rx_station_id)].append(passage)
        else:
            groupped_passages[(tx_station_id, rx_station_id)] = [passage]

    return groupped_passages


def find_passages(
    station_map: dict[StationId, Station],
    station_id_pairs: t.Sequence[tuple[StationId, StationId]],
    space_objects: t.Sequence[sorts.SpaceObject],
    epoch: Datetime_Like,
    spobjs_smpl_dsec: list[npt.NDArray[Float64_as_sec]],
    spobjs_smpl_states: list[EcefStates],
) -> dict[int, list[Passage]]:
    """
    Find passages for each space objects over the simulation period.

    Returns a `list[Passage]` per space object.
    """

    passages_map: dict[int, list[Passage]] = {}

    for spobj_idx, (spobj, spobj_smpl_dsec, spobj_smpl_states) in enumerate(zip(
        space_objects,
        spobjs_smpl_dsec,
        spobjs_smpl_states,
    )):
        passages_of_spobj: list[Passage] = []

        for stn_id_pair in station_id_pairs:
            tx_stn = station_map[stn_id_pair[0]]
            rx_stn = station_map[stn_id_pair[1]]

            passages_of_spobj.extend(
                simulation.funcs.find_passages(
                    dt=spobj_smpl_dsec,
                    space_object=spobj,
                    states=spobj_smpl_states,
                    tx_station=tx_stn,
                    rx_station=rx_stn,
                    epoch=epoch,
                )
            )

        passages_map[spobj_idx] = passages_of_spobj

    return passages_map


# TODO: should be tailored per experiment?
def iter_mpi_simulation_results(save_dir: Path):
    for fpath in save_dir.glob(sim_unit_fname_tpl.format(id="*")):
        with open(fpath, "rb") as f:
            sim_unit: SimulationUnit = pickle.load(f)
            yield sim_unit


# TODO: we need to enforce each station to has a unique id (`.uid` prop)
#   either in the simulation class or in related station getter like `get_radar`
class StxMrxSimulation:
    """
    NOTE: This is intended as an internal constructor, please use the constructor methods to create instances.
    """

    def __init__(
        self,
        station_map: dict[StationId, Station],
        station_id_pairs: t.Sequence[tuple[StationId, StationId]],
        schedule: Schedule,
        exp_detail_map: ExperimentDetailMap,
        epoch: Datetime_Like,
        start_time: Datetime_Like,
        end_time: Datetime_Like,
        space_objects: t.Sequence[sorts.SpaceObject],
        interpolated_propagations: t.Sequence[InterpolatedPropagation],
        passages: dict[int, list[Passage]] | None = None,
        progress: bool = False,
    ):
        self.station_map = station_map
        self.station_id_pairs = station_id_pairs
        self.schedule = schedule
        self.exp_detail_map = exp_detail_map
        self.epoch = epoch
        self.start_time = start_time
        self.end_time = end_time
        self.space_objects = space_objects
        self.interpolated_propagations = interpolated_propagations
        self.progress = progress
        self.passages = passages

        # indexed by space object index in spobj list
        self.sim_units: dict[int, list[SimulationUnit]] = {}
        self.obss: dict[int, list[Observation]] = {}

    @classmethod
    def from_controllers(
        cls,
        controllers: t.Sequence[controller.ControllerBase],
        schedule: Schedule,
        epoch: Datetime_Like,
        start_time: Datetime_Like,
        end_time: Datetime_Like,
        space_objects: t.Sequence[sorts.SpaceObject],
        interpolated_propagations: t.Sequence[InterpolatedPropagation],
        passages: dict[int, list[Passage]] | None = None,
    ):
        """A constructor method"""
        # TODO: - the exp details are already computed outside? Should the `controllers` field be
        # removed? or this classmethod? or what?
        #
        # Notes from Hin, 2025-11-21:
        #   - both `exp_id_stn_id_pairs_map` and `ExperimentDetail` are currently owned by the controller;
        #   - the func `priority_scheduling` evolved to requires `exp_id_stn_id_pairs_map` at some point,
        #     and therefore it is sometimes found as an explicitly variable in simulation experiment file as well
        #   - we can re-work info flow later but this is needed atm

        stn_map: dict[StationId, Station] = {}
        stn_id_pairs_set: set[tuple[StationId, StationId]] = set()
        exp_detail_map: ExperimentDetailMap = {}

        for ctrl in controllers:
            stn_map.update(ctrl.get_station_map())

            for pairs in ctrl.get_experiment_id_station_id_pairs_map().values():
                stn_id_pairs_set.update(pairs)

            exp_detail = ctrl.get_experiment_detail()
            exp_detail_map[exp_detail.id] = exp_detail

        return cls(
            station_map=stn_map,
            station_id_pairs=list(stn_id_pairs_set),
            schedule=schedule,
            exp_detail_map=exp_detail_map,
            epoch=epoch,
            start_time=start_time,
            end_time=end_time,
            space_objects=space_objects,
            interpolated_propagations=interpolated_propagations,
            passages=passages,
        )

    def prepare_simulation_unit_params(
        self,
    ) -> dict[int, list[FromPassagesOverTxRxStationPairParam]]:
        epoch = to_datetime64_us(self.epoch)

        # todo: this is ugly and can be fixed
        dsecs = [
            (interp.times - epoch) / np.timedelta64(1, "s")
            for interp in self.interpolated_propagations
        ]
        states = [interp.states for interp in self.interpolated_propagations]
        if self.passages is None:
            passages_map = find_passages(
                station_map=self.station_map,
                station_id_pairs=self.station_id_pairs,
                space_objects=self.space_objects,
                epoch=self.epoch,
                spobjs_smpl_dsec=dsecs,
                spobjs_smpl_states=states,
            )
            # for obj, objps in enumerate(passages_lists):
            #     print(f"- {obj}")
            #     for ps in objps:
            #         print(f"--  {ps.time_range[0]}")
            logger.debug("find_passages done")
        else:
            passages_map = self.passages

        sim_units_param: dict[int, list[FromPassagesOverTxRxStationPairParam]] = {}
        for spobj_idx, spobj in enumerate(self.space_objects):
            sim_unit_params: list[FromPassagesOverTxRxStationPairParam] = []

            spobj_states_interp = self.interpolated_propagations[spobj_idx].interpolator
            passages_of_a_spobj = passages_map[spobj_idx]
            _SK = scheduling._K

            groupped_passages = group_passages_by_tx_rx_station_pair(passages_of_a_spobj)

            for stn_id_pair, passages in groupped_passages.items():
                tx_stn = self.station_map[stn_id_pair[0]]
                rx_stn = self.station_map[stn_id_pair[1]]
                sched = self.schedule

                filtered_sch = scheduling.filter_by_time_ranges(
                    sched, [ps.time_range for ps in passages]
                )

                # filter by station id
                multi_index_stn_ids_mask = xr.ufuncs.logical_or(
                    filtered_sch[_SK.multi_index][_SK.stn_num] == stn_id_pair[0],
                    filtered_sch[_SK.multi_index][_SK.stn_num] == stn_id_pair[1],
                )
                filtered_sch = filtered_sch[{_SK.multi_index: multi_index_stn_ids_mask}]

                # NOTE: Integers (casted to `str`) are used as `SimulationUnit`s' id
                sim_unit_params.append(
                    FromPassagesOverTxRxStationPairParam(
                        id=str(len(sim_units_param)),
                        passages=passages,
                        spobj=spobj,
                        spobj_interp=spobj_states_interp,
                        tx_station=tx_stn,
                        rx_station=rx_stn,
                        schedule=filtered_sch,
                        exp_detail_map=self.exp_detail_map,
                    )
                )
            sim_unit_params = [
                p for p in sim_unit_params if len(p.schedule[scheduling._K.multi_index]) > 0
            ]
            sim_units_param[spobj_idx] = sim_unit_params

        # filter away param with empty schedule
        logger.debug("prepare_simulation_unit_params done")

        return sim_units_param

    def run(self) -> tuple[dict[int, list[Observation]], dict[int, list[SimulationUnit]]]:
        logger.debug("starting stx mrx sim")

        self.sim_units = {}
        self.obss = {}

        sim_units_param = self.prepare_simulation_unit_params()

        if self.progress:
            pbar = tqdm(desc="simulating", total=len(sim_units_param))

        for spobj_idx, params in sim_units_param.items():
            self.sim_units[spobj_idx] = []
            self.obss[spobj_idx] = []
            for param in params:
                sim_unit = SimulationUnit.from_passages_over_tx_rx_station_pair(param)
                self.sim_units[spobj_idx].append(sim_unit)

                sim_unit.simulate()
                self.obss[spobj_idx].extend(sim_unit.observations)
            if self.progress:
                pbar.update(1)
        if self.progress:
            pbar.close()
        logger.debug("simulation done")

        return self.obss, self.sim_units



#!/usr/bin/env python

""" """
import numpy as np
import pyant

from ..radar import Radar
from ..tx_rx import TX, RX
from ..radar_design import estimate_radar_parameters


def gen_nostra(
    frequency,
    antenna_num,
    antenna_spacing_lambda,
    antenna_efficiency,
    antenna_input_power,
    thermal_load,
    noise_figure_db,
    amplifier_gain_db,
    insertion_loss_db,
    duty_cycle,
    t_sky,
    coherent_integration_time,
    bandwidth_limit_ratio,
):
    """The NOSTRA system."""
    data = estimate_radar_parameters(
        frequency=frequency,
        antenna_num=antenna_num,
        antenna_spacing_lambda=antenna_spacing_lambda,
        antenna_efficiency=antenna_efficiency,
        antenna_input_power=antenna_input_power,
        thermal_load=thermal_load,
        noise_figure_db=noise_figure_db,
        amplifier_gain_db=amplifier_gain_db,
        insertion_loss_db=insertion_loss_db,
        computation_power_draw_scaling=1,
        t_sky=t_sky,
        duty_cycle=duty_cycle,
        reference_snr_db=1,
        reference_ranges=[1000e3],
        coherent_integration_time=coherent_integration_time,
        bandwidth_limit_ratio=bandwidth_limit_ratio,
    )

    dwell_time = coherent_integration_time / duty_cycle
    tx_kw = dict(
        power=data["tx_power"],
        bandwidth=data["effective_bandwidth"],
        duty_cycle=duty_cycle,
        pulse_length=coherent_integration_time / 10,
        ipp=dwell_time / 10,
        n_ipp=10,
        min_elevation=30.0,
    )
    rx_kw = dict(
        noise=data["t_noise"],
        min_elevation=30.0,
    )

    se_rx = RX(
        lat=65.89,
        lon=20.18,
        alt=0,
        beam=data["beam"].copy(),
        beam_parameters=data["beam_parameters"].copy(),
        frequency=frequency,
        **rx_kw,
    )
    se_tx = TX(
        lat=65.89,
        lon=20.18,
        alt=0,
        beam=data["beam"].copy(),
        beam_parameters=data["beam_parameters"].copy(),
        frequency=frequency,
        **tx_kw,
    )

    no_rx = RX(
        lat=68.96,
        lon=18.135,
        alt=0,
        beam=data["beam"].copy(),
        beam_parameters=data["beam_parameters"].copy(),
        frequency=frequency,
        **rx_kw,
    )
    no_tx = TX(
        lat=68.96,
        lon=18.135,
        alt=0,
        beam=data["beam"].copy(),
        beam_parameters=data["beam_parameters"].copy(),
        frequency=frequency,
        **tx_kw,
    )

    fi_rx = RX(
        lat=67.80,
        lon=27.684,
        alt=0,
        beam=data["beam"].copy(),
        beam_parameters=data["beam_parameters"].copy(),
        frequency=frequency,
        **rx_kw,
    )
    fi_tx = TX(
        lat=67.80,
        lon=27.684,
        alt=0,
        beam=data["beam"].copy(),
        beam_parameters=data["beam_parameters"].copy(),
        frequency=frequency,
        **tx_kw,
    )
    # define transmit and receive antennas for a radar network.
    tx = [se_tx, no_tx, fi_tx]
    rx = [se_rx, no_rx, fi_rx]

    nostra = Radar(
        tx=tx,
        rx=rx,
        min_SNRdb=12.0,
        joint_stations=[(0, 0), (1, 1), (2, 2)],
    )
    return nostra



#!/usr/bin/env python

"""Defines a population of space objects in the form of a class."""
from pathlib import Path
from typing import Self, Callable, Any
import copy
from collections import defaultdict, OrderedDict
from functools import reduce

import h5py
import numpy as np
import numpy.typing as npt
import pyorb
from tabulate import tabulate
from astropy.time import Time

from sorts.types import StateType, NDArray_6xN, NDArray_N, IndexLike, AnomalyType, Frames
from sorts.space_object import SpaceObject


class Population:
    """Encapsulates a population of space objects as an array and functions
    for returning instances of space objects.

    The columns represent all components needed to instantiate a space object, i.e. a state (pyorb),
    properties, epoch, and identifiers.
    """

    def __init__(
        self,
        states: NDArray_6xN,
        epochs: Time,
        frame: Frames,
        parameters: dict[str, NDArray_N],
        object_ids: NDArray_N | None,
        state_format: StateType = "kepler",
        anomly_type: AnomalyType = "mean",
        dtypes: dict[str, npt.DTypeLike] | None = None,
        default_dtype: npt.DTypeLike = np.float64,
        epoch_format: str = "mjd",
        epoch_scale: str = "utc",
        degrees: bool = True,
    ):
        self.data: npt.NDArray

        assert states.shape[1] == len(epochs)
        if state_format == "kepler":
            state_keys = pyorb.Orbit.KEPLER
        elif state_format == "cartesian":
            state_keys = pyorb.Orbit.CARTESIAN

        if dtypes is None:
            dtypes = {}

        for dt in list(dtypes.values()):
            if np.dtype(dt).char == "U":
                raise TypeError(
                    "Initialized Population cannot use the save function with"
                    "Unicode [U] numpy strings, try using ASCII [S] strings instead."
                )
        self.state_fields = state_keys
        self.property_fields = list(parameters.keys())
        self.state_format = state_format
        self.epoch_format = epoch_format
        self.epoch_scale = epoch_scale
        self.anomly_type = anomly_type
        self.degrees = degrees
        self.frame = frame

        data_keys = ["id", "epoch"] + state_keys + list(parameters.keys())
        self.dtypes = OrderedDict()
        for key in data_keys:
            dt = dtypes[key] if key in dtypes else default_dtype
            self.dtypes[key] = dt

        self.allocate(len(epochs))
        self.data["id"] = np.arange(len(epochs)) if object_ids is None else object_ids
        self.data["epoch"] = epochs.to_value(epoch_format)
        for ind, key in enumerate(state_keys):
            self.data[key] = states[ind, :]
        for key, vals in parameters.items():
            self.data[key] = vals

    def __len__(self) -> int:
        return len(self.data)

    def copy(self) -> Self:
        """Return a copy of the current Population instance."""
        pop = Population()
        raise NotImplementedError()
        pop.data = self.data.copy()
        return pop

    def delete(self, inds: IndexLike):
        """Remove the rows according to the given indices.
        Supports single index, iterable of indices and slices.
        """
        if isinstance(inds, int):
            inds = [inds]
        elif isinstance(inds, slice):
            _inds = range(self.data.shape[0])
            inds = list(_inds[inds])
        elif not (isinstance(inds, list) or isinstance(inds, np.ndarray)):
            raise Exception("Cannot delete indecies given with type {}".format(type(inds)))

        mask = np.full((self.data.shape[0],), True, dtype=bool)
        for ind in inds:
            mask[ind] = False
        self.data = self.data[mask]

    def filter(self, col: str, fun: Callable[[Any], bool]):
        """Filters the population using a boolean function, keeping true values."""
        if col in self.cols:
            mask = np.full((self.data.shape[0],), True, dtype=bool)
            for row in range(self.data.shape[0]):
                mask[row] = fun(self.data[col][row])
            self.data = self.data[mask]
        else:
            raise Exception("No such column: {}".format(col))

    def unique(self, target_epoch=None, col="id"):
        """Reduces a population by eliminating duplicates with same oid.

        If target_epoch is not given, keep the latest instance found.
        If target_epoch is given, the last instance earlier than the epoch
        is kept, or the first after.
        If col is given, this is the field that will have only unique values
        """
        vmap = defaultdict(list)
        for ii, val in enumerate(self.data[col]):
            vmap[val].append(ii)

        # vmap will become catalogue of entries to delete,
        # so only pop()-ed itmes will remain
        for val in vmap:
            if len(vmap[val]) == 1:
                # Already unique, delete nothing
                vmap[val].pop(0)
                continue

            epochs = self.data["epoch"][vmap[val]]
            order = np.argsort(epochs)[::-1]  # vmap[val][order[0]] is latest
            if target_epoch is None:
                vmap[val].pop(order[0])
                continue
            if np.all(epochs > target_epoch):  # No earlier, pick earliest
                vmap[val].pop(order[-1])
                continue
            ii = np.argmax(epochs[order] < target_epoch)  # Find latest epoch < target
            vmap[val].pop(order[ii])

        # Must delete all items in one swoop, or indices will change under our feet
        deletions = reduce(lambda a, b: a + b, vmap.values())
        self.delete(deletions)

    @property
    def cols(self):
        """The columns"""
        return list(self.dtypes.keys())

    @property
    def keys(self):
        """The columns"""
        return list(self.dtypes.keys())

    @property
    def shape(self):
        """This is the shape of the internal data matrix"""
        shape = (len(self.data), len(self.keys))
        return shape

    def allocate(self, length: int):
        """Allocate the internal data array for assignment of objects.

        **Warning:** This removes all internal data.
        """
        _dtype = []
        for name, dt in self.dtypes.items():
            _dtype.append((name, dt))
        self.data = np.empty((length,), dtype=_dtype)

    def get_states(
        self,
        row_indecies: IndexLike | None = None,
        dtype: npt.DTypeLike | None = None,
    ):
        """Use the defined state parameters to get a copy of the states"""
        return self.get_fields(fields=self.state_fields, row_indecies=row_indecies, dtype=dtype)

    def get_fields(
        self,
        fields: list[str],
        row_indecies: IndexLike | None = None,
        dtype: npt.DTypeLike | None = None,
    ) -> npt.NDArray:
        """Get the orbital elements for one row from internal data array.
        If the `dtype` is not `None`, a structured numpy array is returned,
        otherwise all data from the fields is typecast to the given `dtype`.
        """
        if row_indecies is None:
            row_indecies = slice(None, None, None)  # all

        states = self.data[row_indecies][fields]
        if dtype is not None:
            states_ = np.empty((len(states), len(fields)), dtype=dtype)

            for ind, key in enumerate(fields):
                states_[:, ind] = states[key].astype(dtype)
            states = states_

        return states

    def get_orbit(
        self,
        row_indecies: IndexLike | None = None,
        M_cent: float = pyorb.M_earth,
    ) -> pyorb.Orbit:
        """Get the one row from the population as a :class:`pyorb.Orbit` instance."""
        if row_indecies is None:
            row_indecies = slice(None, None, None)  # all

        fields = self.state_fields
        kwargs = {}
        if isinstance(row_indecies, int) or isinstance(row_indecies, np.integer):
            size = 1
        else:
            size = len(np.arange(len(self.data))[row_indecies])

        for key in fields:
            kwargs[key] = self.data[row_indecies][key]

        obj = pyorb.Orbit(
            M0=M_cent,
            degrees=self.degrees,
            type=self.anomly_type,
            auto_update=True,
            direct_update=True,
            num=size,
            **kwargs,
        )
        return obj

    def get_object(
        self,
        index: int,
        M_cent: float = pyorb.M_earth,
    ) -> SpaceObject:
        """Get the one row from the population as a `SpaceObject` instance."""

        orb = self.get_orbit(index, M_cent=M_cent)
        obj = SpaceObject(
            state=orb,
            frame=self.frame,
            epoch=Time(self.data["epoch"][index], format=self.epoch_format, scale=self.epoch_scale),
            properties={key: self.data[key][index] for key in self.property_fields},
            object_id=self.data["id"][index],
        )
        return obj

    def print(
        self,
        row_indecies: IndexLike | None = None,
        fields: list[str] | None = None,
    ) -> str:
        if row_indecies is None:
            row_indecies = slice(None, None, None)
        if fields is None:
            fields = self.keys

        data = self.data[row_indecies][fields]

        if isinstance(data, np.void):
            data = [[x for x in data]]

        return tabulate(data, headers=fields)

    def __str__(self):
        return self.print()

    def __iter__(self):
        self.__num = 0
        return self

    def __next__(self):
        if self.__num < self.data.shape[0]:
            ret = self.get_object(self.__num)
            self.__num += 1
            return ret
        else:
            raise StopIteration

    @property
    def generator(self):
        for obj in self:
            yield obj

    def save(self, fname: Path | str):
        if isinstance(fname, str):
            fname = Path(fname)

        with h5py.File(fname, "w") as hf:
            hf.create_dataset("data", data=self.data)
            hf.attrs["state_fields"] = self.state_fields
            hf.attrs["property_fields"] = self.property_fields
            hf.attrs["state_format"] = self.state_format
            hf.attrs["epoch_format"] = self.epoch_format
            hf.attrs["epoch_scale"] = self.epoch_scale
            hf.attrs["anomly_type"] = self.anomly_type
            hf.attrs["degrees"] = self.degrees
            hf.attrs["frame"] = self.frame

    @classmethod
    def load(cls, fname: Path | str) -> Self:
        if isinstance(fname, str):
            fname = Path(fname)

        with h5py.File(fname, "r") as hf:
            state_fields = copy.deepcopy(hf.attrs["state_fields"].tolist())
            data = hf["data"][()]
            dtypes = {key: data.dtype[key] for key in data.dtype.names}
            pop = cls(
                states=np.stack([data[key] for key in state_fields]),
                epochs=Time(
                    data["epoch"],
                    format=hf.attrs["epoch_format"],
                    scale=hf.attrs["epoch_scale"],
                ),
                frame=hf.attrs["frame"],
                parameters={
                    key: data[key] for key in copy.deepcopy(hf.attrs["property_fields"].tolist())
                },
                object_ids=data["id"],
                state_format=hf.attrs["state_format"],
                anomly_type=hf.attrs["anomly_type"],
                dtypes=dtypes,
                epoch_format=hf.attrs["epoch_format"],
                epoch_scale=hf.attrs["epoch_scale"],
                degrees=hf.attrs["degrees"],
            )

        return pop



#!/usr/bin/env python

"""Defines what a radar observation schema is in the form of a class.

"""

# Python standard import
from abc import ABC, abstractmethod


# Third party import
import numpy as np

# Local import
from ... import frames


class Scan(ABC):
    """Encapsulates the observation schema of a radar system, i.e. its "scan".

    :param str coordinates: The coordinate system used, can be :code:`'azelr'`,
    :code:`'ned'` or :code:`'enu'`. If `azelr` is used, degrees are assumed.

    **Pointing function:**

    The pointing function must follow the following standard:

     * Take in time in seconds past reference epoch in seconds as first argument.
     * Take any number of keyword arguments.
     * It must return the pointing coordinates as an `(3,)`, `(3,n)` or `(3,n,m)`
        numpy ndarray where `n` is the length of the input time vector and `m`
        is the number of simultaneous pointing directions.
     * Units are in meters.
     * Should be vectorized according to time as the second axis.

    Example pointing function:

    .. code-block:: python

        import numpy as np
        #TODO

    **Coordinate systems:**

     :azelr: Azimuth, Elevation, Range in degrees east of north and above horizon.
     :ned: Cartesian coordinates in North, East, Down.
     :enu: Cartesian coordinates in East, North, Up.


    """

    def __init__(self, coordinates="enu"):
        self.coordinates = coordinates.lower()

    def dwell(self, t):
        """The current dwell time of the scan."""
        return None

    def min_dwell(self):
        """If there are dynamic dwell times, this is the minimum dwell time.
        Otherwise, returns same as :code:`dwell`."""
        return None

    def cycle(self):
        """The cycle time of the scan if applicable."""
        return None

    def copy(self):
        """Return a copy of the current instance."""
        raise NotImplementedError()

    def check_dwell_tx(self, tx):
        """Checks if the transmitting antenna pulse pattern and coherent
        integration schema is compatible with the observation schema.
        Raises an Exception if not.

        :param sorts.radar.TX tx: The antenna that should perform this scan.
        """
        time_slice = tx.n_ipp * tx.ipp
        dwell_time = self.min_dwell()
        if dwell_time is not None:
            if time_slice > dwell_time:
                raise Exception(
                    f"TX time_slice of {time_slice:.2f} s incompatible"
                    f"with minimum dwell time of scan {dwell_time:.2f} s"
                )

    @abstractmethod
    def pointing(self, t):
        pass

    def enu_pointing(self, t):
        """Returns the instantaneous pointing in East, North, Up (ENU) local coordinates.

        :param float/numpy.ndarray t: Seconds past a reference epoch to retrieve the pointing at.
        """
        point = self.pointing(t)

        if self.coordinates == "ned":
            point[2, ...] = -point[2, ...]
        elif self.coordinates == "enu":
            pass
        elif self.coordinates == "azelr":
            if len(point.shape) == 3:
                p_ = np.zeros(point.shape, dtype=point.dtype)
                for ind in range(point.shape[2]):
                    p_[:, :, ind] = frames.sph_to_cart(point[:, :, ind], degrees=True)
                point = p_
                del p_
            else:
                point = frames.sph_to_cart(point, degrees=True)

        return point

    def _transform_ecef(self, point, ant):
        if self.coordinates == "ned":
            k0 = frames.ned_to_ecef(ant.ecef_lat, ant.ecef_lon, ant.ecef_alt, point, degrees=True)
        elif self.coordinates == "enu":
            k0 = frames.enu_to_ecef(ant.ecef_lat, ant.ecef_lon, ant.ecef_alt, point, degrees=True)
        elif self.coordinates == "azelr":
            k0 = frames.azel_to_ecef(
                ant.ecef_lat, ant.ecef_lon, ant.ecef_alt, point[0, ...], point[1, ...], degrees=True
            )
        return k0

    def ecef_pointing(self, t, ant):
        """Returns the instantaneous WGS84 ECEF pointing direction and the radar
        geographical location in WGS84 ECEF coordinates.

        :param float t: Seconds past a reference epoch to retrieve the pointing at.
        """
        point = self.pointing(t)

        if len(point.shape) == 3:
            k0 = np.zeros(point.shape, dtype=point.dtype)
            for ind in range(point.shape[2]):
                k0[:, :, ind] = self._transform_ecef(point[:, :, ind], ant)
        else:
            k0 = self._transform_ecef(point, ant)
        return k0



#!/usr/bin/env python

"""
"""

import pyant

from ..radar import Radar
from ..tx_rx import TX, RX


def gen_tromso_space_debris_radar(fence=False, phased=False):
    lat = 69.5866115
    lon = 19.221555
    alt = 85.0

    if phased:
        rx_beam = pyant.beam_of_radar("tsdr", "phased_fcp")
        tx_beam = pyant.beam_of_radar("tsdr", "phased_fcp")
    else:
        rx_beam = pyant.beam_of_radar("tsdr", "fcp")
        tx_beam = pyant.beam_of_radar("tsdr", "fcp")

    if fence:
        rx_beam.sph_point(azimuth=0.0, elevation=[30.0, 60.0, 90.0, 100.0])
        rx_beam.width = rx_beam.width / 4

        tx_beam.sph_point(azimuth=0.0, elevation=[30.0, 60.0, 90.0, 100.0])
        tx_beam.width = tx_beam.width / 4
    else:
        rx_beam.sph_point(azimuth=0.0, elevation=90.0)
        tx_beam.sph_point(azimuth=0.0, elevation=90.0)

    tsr_tx = TX(
        lat=lat,
        lon=lon,
        alt=alt,
        min_elevation=0,
        beam=tx_beam,
        power=500.0e3,
        bandwidth=1e7,
        duty_cycle=0.125,
        pulse_length=0.0002,
        ipp=10e-3,
        n_ipp=10,
    )

    tsr_rx = RX(
        lat=lat,
        lon=lon,
        alt=alt,
        min_elevation=0,
        noise=100,
        beam=rx_beam,
    )

    tx = [tsr_tx]
    rx = [tsr_rx]

    tsdr_r = Radar(
        tx,
        rx,
        min_SNRdb=10.0,
        joint_stations=[(0, 0)],
    )
    return tsdr_r



#!/usr/bin/env python

"""This module is used to define the radar system"""
import copy
from .. import passes


class Radar(object):
    """A network of transmitting and receiving radar stations.

    :ivar list tx: List of transmitting sites,
        i.e. instances of :class:`sorts.radar.TX`
    :ivar list rx: List of receiving sites,
        i.e. instances of :class:`sorts.radar.RX`
    :ivar float min_SNRdb: Minimum SNR detectable by radar system in dB
        (after coherent integration).
    :ivar list joint_stations: A list of (tx,rx) indecies of stations that
        share hardware. This can be used to e.g. turn of receivers when
        the same hardware is transmitting.

    :param list tx: List of transmitting sites,
        i.e. instances of :class:`sorts.radar.TX`
    :param list rx: List of receiving sites,
        i.e. instances of :class:`sorts.radar.RX`
    :param float min_SNRdb: Minimum SNR detectable by radar system in dB
        (after coherent integration).

    """

    def __init__(self, tx, rx, min_SNRdb=10.0, joint_stations=None):
        self.tx = tx
        self.rx = rx
        self.min_SNRdb = min_SNRdb
        if joint_stations is None:
            self.joint_stations = []
        else:
            self.joint_stations = joint_stations

    def copy(self):
        """Create a deep copy of the radar system."""
        ret = Radar(
            tx=[],
            rx=[],
            min_SNRdb=copy.deepcopy(self.min_SNRdb),
            joint_stations=copy.deepcopy(self.joint_stations),
        )
        for tx in self.tx:
            ret.tx.append(tx.copy())
        for rx in self.rx:
            ret.rx.append(rx.copy())
        return ret

    def set_beam(self, beam):
        """Sets the radiation pattern for transmitters and receivers.

        :param pyant.Beam beam: The radiation pattern to set for radar system.
        """
        self.set_tx_beam(beam)
        self.set_rx_beam(beam)

    def set_tx_beam(self, beam):
        """Sets the radiation pattern for transmitters.

        :param pyant.Beam beam: The radiation pattern to set for radar system.
        """
        for tx in self.tx:
            tx.beam = beam.copy()

    def set_rx_beam(self, beam):
        """Sets the radiation pattern for receivers.

        :param pyant.Beam beam: The radiation pattern to set for radar system.
        """
        for rx in self.rx:
            rx.beam = beam.copy()

    def find_passes(self, t, states, cache_data=True, fov_kw=None):
        """Finds all passes that are simultaneously inside a transmitter
        station FOV and a receiver station FOV.

            :param numpy.ndarray t: Vector of times in seconds to use as a
                base to find passes.
            :param numpy.ndarray states: ECEF states of the object to find
                passes for.
            :return: list of passes indexed by first tx-station and then
                rx-station.
            :rtype: list of list of sorts.Pass
        """
        rd_ps = []
        for txi, tx in enumerate(self.tx):
            rd_ps.append([])
            for rxi, rx in enumerate(self.rx):
                txrx = passes.find_simultaneous_passes(
                    t,
                    states,
                    [tx, rx],
                    cache_data=cache_data,
                    fov_kw=fov_kw,
                )
                for ps in txrx:
                    ps.station_id = [txi, rxi]
                rd_ps[-1].append(txrx)
        return rd_ps



#!/usr/bin/env python

"""Interpolation functions.

"""

# TODO: unit tests for interpolators!!!

from abc import ABC, abstractmethod
import numpy as np
import numpy.typing as npt


class Interpolator(ABC):
    """Base Interpolation class that mimics the behavior of :code:`SpaceObject` so that a `Interpolator` instance can be used instead.

    To create a Interpolator one must define the :code:`get_state` method. to return interpolated
    This method should return states based on the data contained in the instance.
    This data is preferably internalized at instantiation.

    :param numpy.ndarray states: (6,n) array of states to interpolate between.
    :param numpy.ndarray t: (n,) vector of times corresponding to the states.
    """

    def __init__(self, states, t):
        self.states = states
        self.t = np.atleast_1d(t)

    @abstractmethod
    def get_state(self, t) -> npt.NDArray:
        pass


class Legendre8(Interpolator):
    """Order-8 Legendre polynomial interpolation of uniformly distributed states."""

    def __init__(self, states, t):
        if len(t) < 9:
            raise ValueError(f"Cannot performance 8-degree interpolation with {len(t)} points")
        super().__init__(states, t)

    def get_state(self, t):
        intep_states = legendre8(self.states.T, self.t.min(), self.t.max(), t, ti=None)
        return intep_states.T


class Linear(Interpolator):
    """Linear interpolation between states"""

    def __init__(self, states, t):
        super().__init__(states, t)

        self.t_diffs = np.diff(t)

    def get_state(self, t):
        st_t = self.t.flatten()
        in_t = np.atleast_1d(t).flatten()
        # TODO: this can be optimized!
        inds = np.empty(in_t.shape, dtype=np.int64)
        dts = np.empty(in_t.shape, dtype=np.float64)
        for ind in range(len(in_t)):
            t_mat = st_t - in_t[ind]
            inds[ind] = np.argmax(t_mat > 0) - 1
            dts[ind] = -t_mat[inds[ind]]
        frac = dts / self.t_diffs[inds]

        intep_states = self.states[:, inds] * (1 - frac) + self.states[:, inds + 1] * frac
        return intep_states


def legendre8(table, t1, tN, t, ti=None):
    """Order-8 Legendre polynomial interpolation

    Code adapted from the `gdar` system developed by NORCE Norwegian Research Centre AS, used with permission.

    Parameters:
      table: M vectors (M >= 9) to interpolate between, each containing N
          values, e.g. for a position vector N=3 (x, y, z) table.shape = (M, N)
      t1: time corresponding to M=0
      tN: time corresponding to M=N-1
      t: times at which to provide N-dimensional answer.
      ti: indices which sort t in monotonic order

      This version requires t to be sorted, which enables optimization
      by interpolating all t in an interval between two nodes as a single
      expression. For the typical case where there are many fewer intervals
      to consider than distinct values to interpolate, this gives much
      improved performance."""

    M, N = table.shape
    t = np.atleast_1d(t)

    tdiff = t[1:] - t[:-1]
    tzero = 0 * tdiff

    if ti is None:
        if np.any(tdiff < tzero) and np.any(tdiff > tzero):
            # Input must be monotonically increasing or decreasing
            # raise RuntimeError("Input t must be sorted")
            # print("Sorting inputs ...")
            ti = np.argsort(t)
            # print("done")

    if ti is not None:
        t = t[ti]

    rval = np.zeros((len(t), N))

    den = np.array([40320.0, -5040.0, 1440.0, -720.0, 576.0, -720.0, 1440.0, -5040.0, 40320.0])
    trel = (t - t1) / (tN - t1) * (M - 1)
    tind = np.clip(np.round(trel - 4), 0, M - 9).astype(int)

    u, ix = np.unique(tind, return_index=True)
    counts = np.r_[ix[1:], len(tind)] - ix

    err_save = np.seterr(invalid="ignore")
    for val, ix0, count in zip(u, ix, counts):
        six = slice(ix0, ix0 + count)
        xx = (trel[six] - val)[np.newaxis, :] - np.arange(9)[:, np.newaxis]  # {9, count}
        num = np.prod(xx, 0)  # {count}

        rval[six] = np.dot((num[np.newaxis, :] / den[:, np.newaxis] / xx).T, table[val : val + 9])

        zz = np.where(num == 0)[0]
        if len(zz):
            rval[six][zz] = table[np.round(trel[six][zz] - val).astype(int) + val]
    np.seterr(**err_save)

    if ti is not None:
        # Restore input order
        rval, tmp = np.zeros_like(rval), rval
        rval[ti] = tmp
    return rval


def legendre8_loop(table, t1, tN, t):
    """Order-8 Legendre polynomial interpolation

    Code adapted from the `gdar` system developed by NORCE Norwegian Research Centre AS, used with permission.

    Parameters:
      table: M vectors (M >= 9) to interpolate between, each containing N
          values, e.g. for a position vector N=3 (x, y, z) table.shape = (M, N)
      t1: time corresponding to M=0
      tN: time corresponding to M=N-1
      t: times at which to provide N-dimensional answer.

      This version loops over all interpolation instants t.
      Direct port of IDL, which was directly ported from Fortran."""

    # Return value: shape (P, N)
    M, N = table.shape
    t = np.atleast_1d(t)
    P = len(t)

    rval = np.zeros((P, N))

    # denominators are -1^n*n!*(8-n)! for n in [0, ..., 8]
    # (https://oeis.org/A098361)
    den = np.array([40320.0, -5040.0, 1440.0, -720.0, 576.0, -720.0, 1440.0, -5040.0, 40320.0])

    for i in range(P):
        trel = (t[i] - t1) / (tN - t1) * (M - 1)
        tind = np.clip(np.round(trel - 4), 0, M - 9).astype(int)

        x = trel - tind
        num = np.prod(x - np.arange(9))  # x*(x-1)*(x-2)* ... * (x-8)

        if num == 0:
            rval[i] = table[tind + np.rint(x).astype(int)]
        else:
            rval[i] = np.dot(table[tind : tind + 9].T, num / den / (x - np.arange(9)))

    return rval



#!/usr/bin/env python

"""Errors

"""
# Python standard import
import pathlib

# Third party import
import numpy as np
import h5py
import scipy.interpolate
import scipy.constants
import scipy.signal
from tqdm import tqdm

# Local import
from .errors import Errors
from sorts.data import DATA


def simulate_echo(codes, t_vecs, bw=1e6, dop_Hz=0.0, range_m=1e3, sr=5e6):
    """
    Simulate a radar echo with range and Doppler.
    Use windowing to simulate a continuous finite bandwidth signal.
    This is used for linearized error estimates of range and range-rate errors.
    """

    codelen = len(codes[0])
    n_codes = len(codes)
    tvec = np.zeros(codelen * n_codes)
    z = np.zeros(codelen * n_codes, dtype=np.complex64)
    for ci, code in enumerate(codes):
        z[np.arange(codelen) + ci * codelen] = code
        tvec[np.arange(codelen) + ci * codelen] = t_vecs[ci]
    tvec_i = np.copy(tvec)
    tvec_i[0] = tvec[0] - 1e99
    tvec_i[len(tvec) - 1] = tvec[len(tvec) - 1] + 1e99
    zfun = scipy.interpolate.interp1d(tvec_i, z, kind="linear")
    dt = 2.0 * range_m / scipy.constants.c

    z = zfun(tvec + dt) * np.exp(1j * np.pi * 2.0 * dop_Hz * tvec)

    return z


def lin_error(enr=10.0, txlen=1000.0, n_ipp=10, ipp=20e-3, bw=1e6, dr=10.0, ddop=1.0, sr=100e6):
    """
     Determine linearized errors for range and range-rate error
     for a psuedorandom binary phase coded radar transmit pulse
     with a certain transmit bandwidth (inverse of bit length)

     calculate line of sight range and range-rate error,
     given ENR after coherent integration (pulse compression)
     txlen in microseconds.

    Simulate a measurement and do a linearized error estimate.

    """
    codes = []
    t_vecs = []
    n_bits = int(bw * txlen / 1e6)
    oversample = int(sr / bw)
    wfun = scipy.signal.hamming(oversample)
    wfun = wfun / np.sum(wfun)
    for i in range(n_ipp):
        bits = np.array(np.sign(np.random.randn(n_bits)), dtype=np.complex64)
        zcode = np.zeros(n_bits * oversample + 2 * oversample, dtype=np.complex64)
        for j in range(oversample):
            zcode[np.arange(n_bits) * oversample + j + oversample] = bits

        # filter signal so that phase transitions are not too sharp
        zcode = np.convolve(wfun, zcode, mode="same")
        codes.append(zcode)
        tcode = np.arange(n_bits * oversample + 2 * oversample) / sr + float(i) * ipp
        t_vecs.append(tcode)

    z0 = simulate_echo(codes, t_vecs, dop_Hz=0.0, range_m=0.0, bw=bw, sr=sr)
    tau = float(n_ipp) * txlen / 1e6

    # convert coherently integrated ENR to SNR
    # (variance of the measurement errors at receiver bandwidth)
    snr = enr / (tau * sr)

    z_dr = simulate_echo(codes, t_vecs, dop_Hz=0.0, range_m=dr, bw=bw, sr=sr)
    z_diff_r = (z0 - z_dr) / dr

    z_ddop = simulate_echo(codes, t_vecs, dop_Hz=ddop, range_m=0.0, bw=bw, sr=sr)
    z_diff_dop = (z0 - z_ddop) / ddop

    t_l = len(z_dr)
    A = np.zeros([t_l, 2], dtype=np.complex64)
    A[:, 0] = z_diff_r
    A[:, 1] = z_diff_dop
    S = np.real(np.linalg.inv(np.dot(np.transpose(np.conj(A)), A)) / snr)

    return np.sqrt(np.diag(S))


def precalculate_dr(txlen, bw, ipp=20e-3, n_ipp=20, n_interp=20):
    enrs = 10.0 ** np.linspace(-3, 10, num=n_interp)
    drs = np.zeros(n_interp)
    ddops = np.zeros(n_interp)
    q = tqdm(total=n_interp)
    for ei, s in enumerate(enrs):
        dr, ddop = lin_error(enr=s, txlen=txlen, bw=bw, ipp=ipp, n_ipp=n_ipp)
        drs[ei] = dr
        ddops[ei] = ddop
        q.update(1)
        q.set_description(f"Linear errors at SNR {10*np.log10(s)} dB")
    q.close()
    return enrs, drs, ddops


class LinearizedCoded(Errors):
    """
     Determine linearized errors for range and range-rate error
     for a psuedorandom binary phase coded radar transmit pulse
     with a certain transmit bandwidth (inverse of bit length)

     calculate line of sight range and range-rate error,
     given ENR after coherent integration (pulse compression)
     txlen in microseconds.

    Simulate a measurement and do a linearized error estimate.

        Unknown Doppler shift due to ionosphere can be up to 0.1 Hz,
        estimate based on typical GNU Ionospheric tomography receiver phase curves.

    """

    VARIABLES = [
        "range",
        "doppler",
        "range_rate",
    ]

    def __init__(self, tx, seed=None, cache_folder=None, min_range_rate_std=0.1, min_range_std=0.1):
        super().__init__(seed=seed)

        self.min_range_std = min_range_std
        self.min_range_rate_std = min_range_rate_std

        bw = tx.bandwidth
        txlen = tx.pulse_length * 1e6  # pulse length in microseconds
        ipp = tx.ipp
        n_ipp = int(tx.n_ipp)
        self.freq = tx.beam.frequency

        fname = f"MCerr_n{n_ipp}_tx{int(txlen)}_ipp{int(ipp*1e6)}_bw{int(bw)}.h5"

        if cache_folder is None:
            try:
                h = h5py.File(DATA[fname], "r")
                enrs = np.copy(h["enrs"][()])
                drs = np.copy(h["drs"][()])
                ddops = np.copy(h["ddops"][()])
                h.close()
            except KeyError:
                enrs, drs, ddops = precalculate_dr(txlen, bw, ipp=ipp, n_ipp=n_ipp, n_interp=20)
        else:
            if not isinstance(cache_folder, pathlib.Path):
                cache_folder = pathlib.Path(cache_folder)
            pth = cache_folder / fname

            if pth.is_file():
                h = h5py.File(pth, "r")
                enrs = np.copy(h["enrs"][()])
                drs = np.copy(h["drs"][()])
                ddops = np.copy(h["ddops"][()])
                h.close()
            else:
                enrs, drs, ddops = precalculate_dr(txlen, bw, ipp=ipp, n_ipp=n_ipp, n_interp=20)
                ho = h5py.File(pth, "w")
                ho["enrs"] = enrs
                ho["drs"] = drs
                ho["ddops"] = ddops
                ho.close()

        self.rfun = scipy.interpolate.interp1d(np.log10(enrs), np.log10(drs))
        self.dopfun = scipy.interpolate.interp1d(np.log10(enrs), np.log10(ddops))

    def range(self, data, snr):
        """Range in m"""
        self.set_numpy_seed()
        return data + np.random.randn(*data.shape) * self.range_std(snr)

    def doppler(self, data, snr):
        """Doppler shift in Hz"""
        self.set_numpy_seed()
        return data + np.random.randn(*data.shape) * self.doppler_std(snr)

    def range_rate(self, data, snr):
        """Range rate in m/s"""
        self.set_numpy_seed()
        return data + np.random.randn(*data.shape) * self.range_rate_std(snr)

    def range_std(self, snr):
        """The expected standard error of a range measurement in m."""
        dr = 10.0 ** (self.rfun(np.log10(snr)))

        # if object diameter is larger than range error, make it at least as big as target
        if self.min_range_std is not None:
            if len(dr.shape) == 0:
                if dr < self.min_range_std:
                    dr = self.min_range_std
            else:
                dr[dr < self.min_range_std] = self.min_range_std

        return dr

    def doppler_std(self, snr):
        """The expected standard error of a Doppler measurement in Hz."""
        ddop = 10.0 ** (self.dopfun(np.log10(snr)))

        if len(ddop.shape) == 0:
            if ddop < self.min_range_rate_std:
                ddop = self.min_range_rate_std
        else:
            ddop[ddop < self.min_range_rate_std] = self.min_range_rate_std
        return ddop

    def range_rate_std(self, snr):
        """The expected standard error of a range rate measurement in m/s."""
        rr_std = scipy.constants.c * self.doppler_std(snr) / self.freq / 2.0
        return rr_std


class LinearizedCodedIonospheric(LinearizedCoded):
    """233e6 Hz worst case scenario at zenith only!"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        r = np.array([0.0, 1000.0, 1000000.0])
        e = np.array([0.0, 0.1, 100.0]) * 150.0
        self.iono_errfun = scipy.interpolate.interp1d(r, e)

    def range_std(self, path_range, snr):
        """The expected standard error of a range measurement in m."""
        dr = 10.0 ** (self.rfun(np.log10(snr)))

        # add ionospheric error
        dr = np.sqrt(dr**2.0 + self.iono_errfun(path_range / 1e3) ** 2.0)

        if self.min_range_std is not None:
            if len(dr.shape) == 0:
                if dr < self.min_range_std:
                    dr = self.min_range_std
            else:
                dr[dr < self.min_range_std] = self.min_range_std

        return dr

    def range(self, data, snr):
        """Range in m"""
        self.set_numpy_seed()
        return data + np.random.randn(*data.shape) * self.range_std(data, snr)



import logging
from .logging import apply_suggested_config


def setup_function():
    # a workaround to avoid pytest printings and test printings interleaved in the same line
    # https://github.com/pytest-dev/pytest/issues/8574#issuecomment-1806404215
    print()


def test_apply_suggested_config():
    logger = logging.getLogger(__name__)
    apply_suggested_config(logger)

    print(f"this is a print message")
    logger.info(f"this is a info log message")



#!/usr/bin/env python

"""Radar scan plot functions

"""

# Python standard import


# Third party import
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.cm as cm
from mpl_toolkits.mplot3d import Axes3D


# Local import
from . import general


def scan(scan, earth=False, ax=None, max_range=4000e3):
    """Plot a full cycle of the scan pattern.

    :param Scan scan: Scan to plot.
    :param bool earth: Plot the surface of the Earth.
    :param float max_range: The range of the pointing directions.
    :param matplotlib.axes ax: The ax to draw the scan on.
    """
    raise NotImplementedError("")

    min_dwell = scan.min_dwell()
    cycle = scan.cycle()

    t = np.arange(0.0, cycle, min_dwell)

    if ax is None:
        fig = plt.figure(figsize=(15, 15))
        ax = fig.add_subplot(111, projection="3d")
        ax.grid(False)
        ax.view_init(15, 5)
    else:
        fig = None

    if earth:
        general.grid_earth(ax)
    plothelp.draw_radar(ax, SC._lat, SC._lon)

    for i in range(len(t)):
        p0, k0 = SC.antenna_pointing(t[i])

        p1 = p0 + k0 * max_range
        if k0[2] < 0:
            ax.plot([p0[0], p1[0]], [p0[1], p1[1]], [p0[2], p1[2]], alpha=0.5, color="red")
        else:
            ax.plot([p0[0], p1[0]], [p0[1], p1[1]], [p0[2], p1[2]], alpha=0.5, color="green")

    if fig is not None:
        ax.set_xlim(p0[0] - max_range, p0[0] + max_range)
        ax.set_ylim(p0[1] - max_range, p0[1] + max_range)
        ax.set_zlim(p0[2] - max_range, p0[2] + max_range)

    return fig, ax


def plot_radar_scan_movie(SC, earth=False, rotate=False, save_str=""):
    """Create a animation of the scan pattern based on the :code:`_scan_time` and the :code:`_function_data['dwell_time']` variable.

    :param RadarScan SC: Scan to plot.
    :param bool earth: Plot the surface of the Earth.
    :param str save_str: String of path to output movie file. Requers an avalible ffmpeg encoder on the system. If string is empty no movie is saved.
    """
    raise NotImplementedError("")

    if "dwell_time" in SC._function_data:
        dwell_time = n.min(SC._function_data["dwell_time"])
    else:
        dwell_time = 0.05

    if SC._scan_time is None:
        scan_time = dwell_time * 100.0
    else:
        scan_time = SC._scan_time

    t = n.linspace(0.0, scan_time, num=n.round(2 * scan_time / dwell_time))

    fig = plt.figure(figsize=(15, 15))
    ax = fig.add_subplot(111, projection="3d")
    ax.view_init(15, 5)

    def update_text(SC, t):
        return SC.name + ", t=%.4f s" % (t * 1e0,)

    titl = fig.text(0.5, 0.94, update_text(SC, t[0]), size=22, horizontalalignment="center")

    max_range = 4000e3

    p0, k0 = SC.antenna_pointing(0)
    p1 = p0 + k0 * max_range * 0.8

    if earth:
        plothelp.draw_earth_grid(ax)
    else:
        plothelp.draw_earth(ax)
    plothelp.draw_radar(ax, SC._lat, SC._lon)
    if k0[2] < 0:
        beam = ax.plot([p0[0], p1[0]], [p0[1], p1[1]], [p0[2], p1[2]], alpha=0.5, color="red")
    else:
        beam = ax.plot([p0[0], p1[0]], [p0[1], p1[1]], [p0[2], p1[2]], alpha=0.5, color="green")

    ax.set_xlim(p0[0] - max_range, p0[0] + max_range)
    ax.set_ylim(p0[1] - max_range, p0[1] + max_range)
    ax.set_zlim(p0[2] - max_range, p0[2] + max_range)

    interval = scan_time * 1e3 / float(len(t))
    rotations = n.linspace(0.0, 360.0 * 2, num=len(t)) % 360.0

    def update(ti, beam):
        _t = t[ti]
        p0, k0 = SC.antenna_pointing(_t)
        p1 = p0 + k0 * max_range * 0.8
        titl.set_text(update_text(SC, _t))
        beam.set_data([p0[0], p1[0]], [p0[1], p1[1]])
        beam.set_3d_properties([p0[2], p1[2]])
        if k0[2] < 0:
            beam.set_color("red")
        else:
            beam.set_color("green")

        if rotate:
            ax.view_init(15, rotations[ti])

        return (beam,)

    ani = animation.FuncAnimation(
        fig, update, frames=range(len(t)), fargs=(beam), interval=interval, blit=False
    )

    if len(save_str) > 0:
        # Set up formatting for the movie files
        Writer = animation.writers["ffmpeg"]
        writer = Writer(metadata=dict(artist="Daniel Kastinen"), bitrate=1800)
        ani.save(save_str, writer=writer)

    plt.tight_layout()
    plt.show()



#!/usr/bin/env python

"""A parent class used for interfacing any propagator."""

# Python standard import
from typing import Generic
import logging
from abc import ABC, abstractmethod

from astropy.time import Time, TimeDelta
from sorts.types import S, NDArray_N, NDArray_6xN
from sorts.utils import convert_to_relative_time
from sorts.space_object import SpaceObject

logger = logging.getLogger(__name__)


class Propagator(ABC, Generic[S]):
    def __init__(self, settings: S):
        self.settings = settings
        for key in self.settings.keys:
            logger.debug(f"Propagator:settings:{key} = {getattr(self.settings, key)}")

    def propagate_to_new_epoch(
        self,
        space_object: SpaceObject,
        dt: Time | TimeDelta | float,
        copy: bool = True,
    ) -> SpaceObject:
        """Propagate and change the epoch of this space object if the state is a `pyorb.Orbit`."""
        dt = convert_to_relative_time(space_object.epoch, dt)[0]
        new_cart = self.propagate(space_object, dt)
        if len(new_cart.shape) < 2:
            new_cart.shape = (new_cart.size, 1)
        obj = space_object.copy() if copy else space_object
        obj.state.cartesian = new_cart
        obj.state.calculate_kepler()
        obj.epoch += TimeDelta(dt, format="sec")
        return obj

    @abstractmethod
    def propagate(
        self,
        space_object: SpaceObject,
        times: Time | TimeDelta | NDArray_N,
    ) -> NDArray_6xN:
        """Propagate a state

        This function uses key-word argument to supply additional information
        to the propagator, such as area or mass.

        The coordinate frames used should be documented in the child class docstring.

        SI units are assumed unless implementation states otherwise.
        """
        pass



#!/usr/bin/env python

"""Defines a space object. Encapsulates orbital elements, propagation and related methods."""
from copy import deepcopy
from typing import Any, Self
from dataclasses import dataclass, fields

import numpy as np
from pyorb import Orbit, M_earth
from astropy.time import Time, TimeDelta
from sorts.types import Frames, NDArray_N


@dataclass
class SpaceObject:
    """Encapsulates a object in space which has a state, at an epoch and in a frame, and some
    properties.

    The state of the object is stored in a `pyorb.Orbit` instance.
    This instance contains direct transformations between the Cartesian and Kepler states.

    To propagate this object in time and get states, supply it to a propagator.
    """

    state: Orbit
    frame: Frames
    epoch: Time
    properties: dict[str, Any]
    object_id: int = 0

    def copy(self) -> Self:
        kwargs = {key: deepcopy(getattr(self, key)) for key in self.keys}
        return self.__class__(**kwargs)

    @property
    def keys(self) -> list[str]:
        return [key.name for key in fields(self)]

    @classmethod
    def from_tle(cls):
        raise NotImplementedError("todo")

    @classmethod
    def from_kepler(
        cls,
        semi_major_axis: float,
        eccentricity: float,
        inclination: float,
        argument_of_periapsis: float,
        longitude_of_ascending_node: float,
        mean_anomaly: float,
        epoch: Time,
        frame: Frames,
        properties: dict[str, Any],
        center_mass: float = M_earth,
        object_id: int = 0,
        degrees: bool = True,
    ):
        state = Orbit(
            M0=center_mass,
            degrees=degrees,
            type="mean",
            auto_update=True,
            direct_update=True,
            num=1,
            m=0.0,
        )
        state._kep[0, 0] = semi_major_axis
        state._kep[1, 0] = eccentricity
        state._kep[2, 0] = inclination
        state._kep[3, 0] = argument_of_periapsis
        state._kep[4, 0] = longitude_of_ascending_node
        state._kep[5, 0] = mean_anomaly
        state.calculate_cartesian()

        return cls(
            state=state,
            frame=frame,
            epoch=epoch,
            properties=properties,
            object_id=object_id,
        )

    def __repr__(self):
        return f"SpaceObject(oid={self.object_id} @ {self.epoch.iso})"

    @property
    def d(self) -> float:
        if "d" in self.properties:
            diam = self.properties["d"]
        elif "diam" in self.properties:
            diam = self.properties["diam"]
        elif "r" in self.properties:
            diam = self.properties["r"] * 2
        elif "A" in self.properties:
            diam = np.sqrt(self.properties["A"] / np.pi) * 2
        else:
            raise AttributeError(
                "Space object does not have a diameter parameter or any way to calculate one"
            )
        return diam

    def __str__(self):
        p = f"\nSpace object {self.object_id}: {repr(self.epoch)}:\n"
        p += str(self.state) + "\n"
        p += "Parameters: " + ", ".join([f"{key}={val}" for key, val in self.properties.items()])
        return p



#!/usr/bin/env python

"""package data

"""
import importlib

DATA = {}

# To be compatible with 3.7-8
# as resources.files was introduced in 3.9
if hasattr(importlib.resources, "files"):
    _data_folder = importlib.resources.files("sorts.data")
    for file in _data_folder.iterdir():
        if not file.is_file():
            continue
        if file.name.endswith(".py"):
            continue

        DATA[file.name] = file

else:
    _data_folder = importlib.resources.contents("sorts.data")
    for fname in _data_folder:
        with importlib.resources.path("sorts.data", fname) as file:
            if not file.is_file():
                continue
            if file.name.endswith(".py"):
                continue

            DATA[file.name] = pathlib.Path(str(file))



#!/usr/bin/env python

"""Plotting sub package

"""

from .general import grid_earth, transformed_grid_earth, set_axes_equal
from .radar import radar_earth, radar_map
from .tracking import local_tracking, local_passes
from .schedule import schedule_pointing, observed_parameters
from .orbits import kepler_scatter, kepler_orbit
from . import colors



#!/usr/bin/env python

""" """
# Python standard import


import numpy as np
import pyant

from .radars import radar_generator
from ..radar import Radar
from ..tx_rx import TX, RX


@radar_generator("eiscat_esr", "cassegrain")
def gen_eiscat_esr():
    # TODO: Proper ESR coordinates.  These were eyeballed from maps
    lat_esr32 = 78.153145
    lon_esr32 = 16.0758715
    alt_esr = 185  # TODO: Check!

    # QnD calculations: 42m is 128 m to the East of the 32m
    # (Computing ECEF coordinates from this gives an offset of 128.66 metres)
    baseline = 128
    lon_incr = baseline / (1e7 / 90 * np.cos(np.radians(lat_esr32)))
    lat_esr42 = lat_esr32
    lon_esr42 = lon_esr32 + lon_incr

    tx_beam32 = pyant.beam_of_radar("esr_32m", "cassegrain")
    rx_beam32 = pyant.beam_of_radar("esr_32m", "cassegrain")
    tx_beam42 = pyant.beam_of_radar("esr_42m", "cassegrain")
    rx_beam42 = pyant.beam_of_radar("esr_42m", "cassegrain")

    tx = [
        TX(  # 1st TX: 32m
            lat=lat_esr32,
            lon=lon_esr32,
            alt=alt_esr,
            min_elevation=15,  # TODO: Check
            beam=tx_beam32,
            power=1.0e6,
            bandwidth=2e6,  # Check
            duty_cycle=0.25,
            pulse_length=2e-3,  # semantics?
            ipp=20e-3,
            n_ipp=10,
        ),
        TX(  # 2nd TX: 42m
            lat=lat_esr42,
            lon=lon_esr42,
            alt=alt_esr,
            min_elevation=75,  # TODO: Fix az and el to narrow intervals
            beam=tx_beam42,
            power=1.0e6,
            bandwidth=2e6,  # Check
            duty_cycle=0.25,
            pulse_length=2e-3,  # semantics?
            ipp=20e-3,
            n_ipp=10,
        ),
    ]

    rx = [
        RX(
            lat=lat_esr32,
            lon=lon_esr32,
            alt=alt_esr,
            min_elevation=15,  # TODO: Check
            noise=70,
            beam=rx_beam32,
        ),
        RX(
            lat=lat_esr42,
            lon=lon_esr42,
            alt=alt_esr,
            min_elevation=75,  # TODO: Fix az and el to narrow intervals
            noise=70,
            beam=rx_beam42,
        ),
    ]

    esr = Radar(
        tx,
        rx,
        min_SNRdb=10.0,
        joint_stations=[(0, 0), (1, 1)],
    )
    return esr



#!/usr/bin/env python

"""Encapsulates a fundamental component of tracking space objects:
a pass over a geographic location.
Also provides convenience functions for finding passes given states
and stations and sorting structures of passes in particular ways.

"""
import datetime

import numpy as np
import numpy.typing as npt
import spacecoords
import pyorb
from sorts.types import Float64_as_sec

# Local import
from .signals import hard_target_snr


class Pass:
    """Saves the local coordinate data for a single pass.
    Optionally also indicates the location of that pass in a bigger dataset.
    """

    def __init__(self, t, enu, inds=None, cache=True, station_id: int | list[int] = 0):
        self.inds = inds  # refer back to the population index
        self.t = t  # time
        self.enu: list[npt.NDArray] | npt.NDArray = enu  # of shapes: (3, n) | ((3, n), ..., k)
        self.cache = cache

        self.station_id = station_id

        self.snr = None

        self._start = None
        self._end = None
        self._r = None
        self._v = None
        self._zang = None

    def __str__(self):
        str_ = "Pass "
        if self.station_id is not None:
            str_ += f"Station {self.station_id} | "
        str_ += (
            f"Rise {str(datetime.timedelta(seconds=self.start()))}"
            f"({(self.end() - self.start())/60.0:.1f} min)"
            f"{str(datetime.timedelta(seconds=self.end()))} Fall"
        )

        return str_

    def __repr__(self):
        return str(self)

    def calculate_snr(self, tx, rx, diameter):
        """Uses the :code:`signals.hard_target_snr` function to calculate the optimal SNR curve
        of a target during the pass **if the TX and RX stations are pointing at the object**.
        The SNR's are returned from the function but also stored in the property :code:`self.snr`.

        :param TX tx: The TX station that observed the pass.
        :param RX rx: The RX station that observed the pass.
        :param float diameter: The diameter of the object.
        :return: (n,) Vector with the SNR's during the pass.
        :rtype: numpy.ndarray

        """
        assert self.stations == 2, "Can only calculate SNR for TX-RX pairs"

        ranges = self.range()
        enus = self.enu

        self.snr = np.empty((len(self.t),), dtype=np.float64)
        for ti in range(len(self.t)):
            tx.beam.point(enus[0][:3, ti])
            rx.beam.point(enus[1][:3, ti])

            self.snr[ti] = hard_target_snr(
                tx.beam.gain(enus[0][:3, ti]),
                rx.beam.gain(enus[1][:3, ti]),
                rx.wavelength,
                tx.power,
                ranges[0][ti],
                ranges[1][ti],
                diameter=diameter,
                bandwidth=tx.coh_int_bandwidth,
                rx_noise_temp=rx.noise,
            )
        return self.snr

    @property
    def stations(self):
        """The number of stations that can observe the pass."""
        if self.station_id is not None:
            if isinstance(self.station_id, list):
                return len(self.station_id)
            else:
                return 1
        else:
            return 1

    def start(self):
        """The Start time of the pass
        (uses cached value after first call if :code:`self.cache=True`)."""
        if self.cache:
            if self._start is None:
                self._start = self.t.min()
            return self._start
        else:
            return self.t.min()

    def end(self):
        """The ending time of the pass
        (uses cached value after first call if :code:`self.cache=True`)."""
        if self.cache:
            if self._end is None:
                self._end = self.t.max()
            return self._end
        else:
            return self.t.max()

    @staticmethod
    def calculate_range(enu):
        """Norm of the ENU coordinates."""
        return np.linalg.norm(enu[:3, :], axis=0)

    @staticmethod
    def calculate_range_rate(enu):
        """Projected ENU velocity along the ENU range."""
        return np.sum(enu[3:, :] * (enu[:3, :] / np.linalg.norm(enu[:3, :], axis=0)), axis=0)

    @staticmethod
    def calculate_zenith_angle(enu, radians=False):
        """Zenith angle of the ENU coordinates."""
        return spacecoords.linalg.vector_angle(
            np.array([0, 0, 1], dtype=np.float64), enu[:3, :], degrees=not radians
        )

    def get_range(self):
        """Get ranges from all stations involved in the pass.

        :rtype: list of numpy.ndarray or numpy.ndarray
        :return: If there is one station observing the pass, return a vector of ranges,
        otherwise return a list of vectors with ranges.
        """
        if self.stations > 1:
            return [Pass.calculate_range(enu) for enu in self.enu]
        else:
            return Pass.calculate_range(self.enu)

    def range(self):
        """Get ranges from all stations involved in the pass
        (uses cached value after first call if :code:`self.cache=True`).
        The cache is stored in :code:`self._r`.

        :rtype: list of numpy.ndarray or numpy.ndarray
        :return: If there is one station observing the pass, return a vector of ranges,
            otherwise return a list of vectors with ranges.
        """
        if self.cache:
            if self._r is None:
                self._r = self.get_range()
            return self._r
        else:
            return self.get_range()

    def get_range_rate(self):
        """Get range rates from all stations involved in the pass.

        :rtype: list of numpy.ndarray or numpy.ndarray
        :return: If there is one station observing the pass, return a vector of range rates,
            otherwise return a list of vectors with range rates.
        """
        if self.stations > 1:
            return [Pass.calculate_range_rate(enu) for enu in self.enu]
        else:
            return Pass.calculate_range_rate(self.enu)

    def range_rate(self):
        """Get range rates from all stations involved in the pass
        (uses cached value after first call if :code:`self.cache=True`).
        The cache is stored in :code:`self._v`.

        :rtype: list of numpy.ndarray or numpy.ndarray
        :return: If there is one station observing the pass, return a vector of range rates,
            otherwise return a list of vectors with range rates.
        """
        if self.cache:
            if self._v is None:
                self._v = self.get_range_rate()
            return self._v
        else:
            return self.get_range_rate()

    def get_zenith_angle(self, radians=False):
        """Get zenith angles for all stations involved in the pass.

        :rtype: list of numpy.ndarray or numpy.ndarray
        :return: If there is one station observing the pass, return a vector of zenith angles,
            otherwise return a list of vectors with zenith angles.
        """
        if self.stations > 1:
            return [Pass.calculate_zenith_angle(enu, radians=radians) for enu in self.enu]
        else:
            return Pass.calculate_zenith_angle(self.enu, radians=radians)

    def zenith_angle(self, radians=False):
        """Get zenith angles from all stations involved in the pass
        (uses cached value after first call if :code:`self.cache=True`).
        The cache is stored in :code:`self._v`.

        :rtype: list of numpy.ndarray or numpy.ndarray
        :return: If there is one station observing the pass, return a vector of zenith angles,
            otherwise return a list of vectors with zenith angles.
        """
        if self.cache:
            if self._zang is None:
                self._zang = self.get_zenith_angle(radians=radians)
            return self._zang
        else:
            return self.get_zenith_angle(radians=radians)


def equidistant_sampling(
    orbit, start_t, end_t, max_dpos=1e3, eccentricity_tol=0.3
) -> npt.NDArray[Float64_as_sec]:
    """Find the temporal sampling of an orbit which is sufficient to achieve a
    maximum spatial separation. Assume elliptic orbit and uses Keplerian propagation
    to find sampling, does not take perturbation patterns into account. If
    eccentricity is small, uses periapsis speed and uniform sampling in time.

    :param pyorb.Orbit orbit: Orbit to find temporal sampling of.
    :param float start_t: Start time in seconds
    :param float end_t: End time in seconds
    :param float max_dpos: Maximum separation between evaluation points in meters.
    :param float eccentricity_tol: Minimum eccentricity below which the orbit is
        approximated as a circle and temporal samples are uniform in time.
    :return: Vector of sample times in seconds.
    :rtype: numpy.ndarray
    """
    if len(orbit) > 1:
        raise ValueError(f"Cannot use vectorized orbits: len(orbit) = {len(orbit)}")

    if orbit.e <= eccentricity_tol:
        r = pyorb.elliptic_radius(0.0, orbit.a, orbit.e, degrees=False)
        v = pyorb.orbital_speed(r, orbit.a, orbit.G * (orbit.M0 + orbit.m))[0]
        return np.arange(start_t, end_t, max_dpos / v)

    tmp_orb = orbit.copy()
    tmp_orb.auto_update = False

    tmp_orb.propagate(start_t)
    period = tmp_orb.period

    t_curr = start_t
    t = [t_curr]
    t_repeat = None
    while t_curr < end_t:
        if t_curr - start_t > period:
            if t_repeat is None:
                t_repeat = len(t)
            dt = t[-t_repeat + 1] - t[-t_repeat]
            t_curr += dt
        else:
            v = tmp_orb.speed[0]
            dt = max_dpos / v
            t_curr += dt
            tmp_orb.propagate(dt)

        t.append(t_curr)
    return np.array(t, dtype=np.float64)


def find_passes(t, states, station, cache_data=True):
    """Find passes inside the FOV of a radar station given a series of times for a space object.

    :param numpy.ndarray t: Vector of times in seconds to use as a base to find passes.
    :param numpy.ndarray states: ECEF states of the object to find passes for.
    :param sorts.Station station: Radar station that defines the FOV, or a list of radar stations.
    :return: list of passes
    :rtype: sorts.Pass

    """
    passes = []

    if isinstance(station, list):
        check = np.logical_or.reduce([st.field_of_view(states) for st in station])
    else:
        check = station.field_of_view(states)
    inds = np.where(check)[0]

    if len(inds) == 0:
        return passes

    dind = np.diff(inds)
    splits = np.where(dind > 1)[0]

    splits = np.insert(splits, 0, -1)
    splits = np.insert(splits, len(splits), len(inds) - 1)
    splits += 1
    for si in range(len(splits) - 1):
        ps_inds = inds[splits[si] : splits[si + 1]]

        if cache_data:
            enu = station.enu(states[:3, :])
            ps = Pass(
                t=t[ps_inds],
                enu=enu[:, ps_inds],
                inds=ps_inds,
                cache=True,
            )
        else:
            ps = Pass(
                t=None,
                enu=None,
                inds=ps_inds,
                cache=True,
            )
            ps._start = t[ps_inds].min()
            ps._end = t[ps_inds].max()

        passes.append(ps)

    return passes


def find_simultaneous_passes(t, states, stations, cache_data=True, fov_kw=None):
    """Finds all passes that are simultaneously inside a multiple stations FOV's.

    :param numpy.ndarray t: Vector of times in seconds to use as a base to find passes.
    :param numpy.ndarray states: ECEF states of the object to find passes for.
    :param list of sorts.Station stations: Radar stations that defines the FOV's.
    :return: list of passes
    :rtype: list of sorts.Pass

    """
    passes = []
    if fov_kw is None:
        fov_kw = {}

    enu = []
    check = np.full((len(t),), True, dtype=bool)
    for station in stations:
        enu_st = station.enu(states)
        enu.append(enu_st)

        check_st = station.field_of_view(states, **fov_kw)
        check = np.logical_and(check, check_st)

    inds = np.where(check)[0]

    if len(inds) == 0:
        return passes

    dind = np.diff(inds)
    splits = np.where(dind > 1)[0]

    splits = np.insert(splits, 0, -1)
    splits = np.insert(splits, len(splits), len(inds) - 1)
    splits += 1
    for si in range(len(splits) - 1):
        ps_inds = inds[splits[si] : splits[si + 1]]
        if len(ps_inds) == 0:
            continue
        if cache_data:
            ps = Pass(
                t=t[ps_inds],
                enu=[xv[:, ps_inds] for xv in enu],
                inds=ps_inds,
                cache=True,
                station_id=[None, None],
            )
        else:
            ps = Pass(
                t=None,
                enu=None,
                inds=ps_inds,
                cache=True,
                station_id=[None, None],
            )
            ps._start = t[ps_inds].min()
            ps._end = t[ps_inds].max()

        passes.append(ps)

    return passes


def group_passes(passes):
    """Takes a list of passes structured as
    [tx][rx][pass] and find all simultaneous passes
    and groups them according to [tx], resulting in a [tx][pass][rx] structure.
    """

    def overlap(ps1, ps2):
        return ps1.start() <= ps2.end() and ps2.start() <= ps1.end()

    grouped_passes = []
    for tx_passes in passes:
        grouped_passes.append([])

        # first flatten
        flat_passes = [x for rx_passes in tx_passes for x in rx_passes]

        if len(flat_passes) > 0:
            grouped_passes[-1].append([flat_passes[0]])
        else:
            continue

        for x in range(1, len(flat_passes)):
            for y in range(len(grouped_passes[-1])):
                member = False
                for gps in grouped_passes[-1][y]:
                    if overlap(gps, flat_passes[x]):
                        member = True
                        break

                if member:
                    member_id = y
                    break

            if member:
                grouped_passes[-1][member_id].append(flat_passes[x])
            else:
                grouped_passes[-1].append([flat_passes[x]])

    return grouped_passes



#!/usr/bin/env python

"""SORTS package"""
import ctypes
import pathlib
import sysconfig
import logging
from .version import __version__

# Get and config module-level logger
logger = logging.getLogger(__name__)

# Find suffix
# suffix = sysconfig.get_config_var("EXT_SUFFIX")
# if suffix is None:
#     suffix = ".so"
#
# __sortspath__ = pathlib.Path(__file__).resolve().parent
# __libpath__ = __sortspath__ / ("clibsorts" + suffix)
#
# clibsorts = ctypes.cdll.LoadLibrary(str(__libpath__))

##
# v2 imports
##
from . import types
from . import utils
from . import scheduling
from . import controller
from . import simulation

from .scheduling import Schedule, ExperimentDetail
from .controller import TrackerController, FenceScanController
from .simulation import StxMrxSimulation
from .mpi_queued_execution import MpiQueuedExecution


##
# v1 imports
##
# TODO: clean up these imports

# modules
from .radar import scans
from . import radar
from . import functions
from . import constants
from . import frames
from . import dates
from . import plotting

# from . import controller_v1
from . import scheduler
from . import passes
from . import errors
from . import io
from . import interpolation

# from . import simulation_v1
from . import signals
from . import correlator
from . import propagator

# classes
from .space_object import SpaceObject
from .population import Population
from .propagator import Propagator
from .radar import Scan
from .radar import Station, TX, RX
from .controller_v1 import RadarController
from .scheduler import Scheduler
from .passes import Pass
from .errors import Errors
from .simulation_v1 import Simulation

# Functions
from .radar import get_radar, list_radars
from .correlator import correlate
from .passes import equidistant_sampling
# from .passes import find_passes, find_simultaneous_passes, group_passes
from .signals import hard_target_snr
from .simulation_v1 import (
    MPI_single_process,
    MPI_action,
    iterable_step,
    store_step,
    cached_step,
    iterable_cache,
    pre_post_actions,
)
from .pre_encounter import propagate_pre_encounter, distance_termination



#!/usr/bin/env python

"""Estimating orbit determination errors

"""
import pathlib

import numpy as np

from .linearized_coded import LinearizedCodedIonospheric
from .linearized_coded import LinearizedCoded


def orbit_determination_covariance(
    passes,
    scheduler,
    space_object,
    variables=["x", "y", "z", "vx", "vy", "vz", "A"],
    deltas=[1e-4] * 3 + [1e-6] * 3 + [1e-2],
    cache_folder=None,
    ray_bending=True,
    prior_cov_inv=None,
    transforms={
        "A": (lambda A: np.log10(A), lambda Ainv: 10.0**Ainv),
    },
):
    """Takes a series of passes and calculates a orbit determination covariance of the measurement model is linear and linearized coded errors with optionally ionospheric effects are included."""

    if cache_folder is not None:
        cache_folder = pathlib.Path(cache_folder)

    # observe all the passes, including measurement Jacobian
    datas = []
    J = None
    for ind, ps in enumerate(passes):
        txi, rxi = ps.station_id

        # Now we load the error model
        if ray_bending:
            err = LinearizedCodedIonospheric(scheduler.radar.tx[txi], cache_folder=cache_folder)
        else:
            err = LinearizedCoded(scheduler.radar.tx[txi], cache_folder=cache_folder)

        # the Jacobean is stacked as [r_measurements, v_measurements]^T so we stack the measurement covariance equally
        data, J_rx = scheduler.calculate_observation_jacobian(
            ps,
            space_object=space_object,
            variables=variables,
            deltas=deltas,
            snr_limit=True,
            transforms=transforms,
        )
        if data is None:
            continue

        datas += [data]

        # now we get the expected standard deviations
        if ray_bending:
            r_stds_tx = err.range_std(data["range"], data["snr"])
        else:
            r_stds_tx = err.range_std(data["snr"])
        v_stds_tx = err.range_rate_std(data["snr"])

        # Assume uncorrelated errors = diagonal covariance matrix
        Sigma_m_diag_tx = np.r_[r_stds_tx**2, v_stds_tx**2]

        # we simply append the results on top of each other for each station
        if J is not None:
            J = np.append(J, J_rx, axis=0)
            Sigma_m_diag = np.append(Sigma_m_diag, Sigma_m_diag_tx, axis=0)
        else:
            J = J_rx
            Sigma_m_diag = Sigma_m_diag_tx

    # This means that no passes were observable
    if J is None:
        return None, None

    # diagonal matrix inverse is just element wise inverse of the diagonal
    Sigma_m_inv = np.diag(1.0 / Sigma_m_diag)

    # For a thorough derivation of this formula:
    # see Fisher Information Matrix of a MLE with Gaussian errors and a Linearized measurement model
    if prior_cov_inv is not None:
        Sigma_orb = np.linalg.inv(np.transpose(J) @ Sigma_m_inv @ J + prior_cov_inv)
    else:
        Sigma_orb = np.linalg.inv(np.transpose(J) @ Sigma_m_inv @ J)

    return Sigma_orb, datas


def covariance_propagation(
    space_object,
    orbit_cov,
    t,
    variables,
    samples=100,
    perturbation_cov=None,
    perturbed_variables=None,
    transforms={
        "A": (lambda A: np.log10(A), lambda Ainv: 10.0**Ainv),
    },
):
    """
    Propagate error covariance in time. The time vector should start at the moment of the epoch for the covariance matrix.
    Sample mean position and velocity error.
    Optionally add additional errors.
    """

    ecef0 = space_object.get_state(t)

    r_diff = np.zeros(len(t), dtype=np.float64)
    v_diff = np.zeros(len(t), dtype=np.float64)
    for i in range(samples):
        deltas = np.random.multivariate_normal(np.zeros(orbit_cov.shape[0]), orbit_cov)

        if perturbation_cov is not None:
            pert = np.random.multivariate_normal(
                np.zeros(perturbation_cov.shape[0]), perturbation_cov
            )

        dso = space_object.copy()

        update = {}
        for ind, var in enumerate(variables):
            if var in transforms:
                Tx = transforms[var][0](getattr(dso, var)) + deltas[ind]
                dx = transforms[var][1](Tx)
            else:
                dx = getattr(dso, var) + deltas[ind]

            update[var] = dx

        if perturbation_cov is not None:
            for ind, var in enumerate(perturbed_variables):
                if var in update:
                    base = update[var]
                else:
                    base = getattr(dso, var)

                if var in transforms:
                    Tx = transforms[var][0](base) + pert[ind]
                    dx = transforms[var][1](Tx)
                else:
                    dx = base + pert[ind]

            update[var] = dx

        dso.update(**update)

        ecef1 = dso.get_state(t)

        r_diff += np.linalg.norm(ecef1[3:, :] - ecef0[3:, :], axis=0)
        v_diff += np.linalg.norm(ecef1[3:, :] - ecef0[3:, :], axis=0)

    r_diff_stdev = r_diff / samples
    v_diff_stdev = v_diff / samples

    return r_diff_stdev, v_diff_stdev



#!/usr/bin/env python

"""
The EISCAT 3D system.

The EISCAT 3D system in stage 1.

For more information see:
    * `EISCAT <https://eiscat.se/>`_
    * `EISCAT 3D <https://www.eiscat.se/eiscat3d/>`_


# TODO update these

**EISCAT 3D Stages:**

    * Stage 1: x
    * Stage 2: x
    * Stage 3: x



"""
import pyant

from .radars import radar_generator
from ..radar import Radar
from ..tx_rx import TX, RX


SKI_LAT = 69.34023844
SKI_LON = 20.313166
SKI_ALT = 0.0

KAR_LAT = 68.463862
KAR_LON = 22.458859
KAR_ALT = 0.0

KAI_LAT = 68.148205
KAI_LON = 19.769894
KAI_ALT = 0.0

KIR_LAT = 67.860308
KIR_LON = 20.432841
KIR_ALT = 300.0


def gen_eiscat3d_stage1(beam_type, path=None, **interp_kw):
    """ """
    if beam_type == "array":
        tx_beam_ski = pyant.beam_of_radar("e3d_stage1", "array")
        rx_beam_ski = pyant.beam_of_radar("e3d_stage2", "array")
        rx_beam_kar = pyant.beam_of_radar("e3d_stage1", "array")
        rx_beam_kai = pyant.beam_of_radar("e3d_stage1", "array")
    if beam_type == "interp":
        tx_beam_ski = pyant.beam_of_radar("e3d_stage1", "interpolated_array", **interp_kw)
        rx_beam_ski = pyant.beam_of_radar("e3d_stage2", "interpolated_array", **interp_kw)
        rx_beam_kar = pyant.beam_of_radar("e3d_stage1", "interpolated_array", **interp_kw)
        rx_beam_kai = pyant.beam_of_radar("e3d_stage1", "interpolated_array", **interp_kw)
    rx_kw = dict(
        min_elevation=30.0,
        noise=150,
    )
    ski = RX(lat=SKI_LAT, lon=SKI_LON, alt=SKI_ALT, beam=rx_beam_ski, **rx_kw)
    dwell_time = 0.1
    ski_tx = TX(
        lat=SKI_LAT,
        lon=SKI_LON,
        alt=SKI_ALT,
        min_elevation=30.0,
        beam=tx_beam_ski,
        power=5e6,  # 5 MW
        bandwidth=100e3,  # 100 kHz tx bandwidth
        duty_cycle=0.25,  # 25% duty-cycle
        pulse_length=1920e-6,
        ipp=10e-3,
        n_ipp=int(dwell_time / 10e-3),
    )
    kar = RX(lat=KAR_LAT, lon=KAR_LON, alt=KAR_ALT, beam=rx_beam_kar, **rx_kw)
    kai = RX(lat=KAI_LAT, lon=KAI_LON, alt=KAI_ALT, beam=rx_beam_kai, **rx_kw)
    # define transmit and receive antennas for a radar network.
    tx = [ski_tx]
    rx = [ski, kar, kai]

    eiscat3d = Radar(
        tx=tx,
        rx=rx,
        min_SNRdb=10.0,
        joint_stations=[(0, 0)],
    )
    return eiscat3d


@radar_generator("eiscat3d", "stage1-array")
def gen_eiscat3d_stage1_array(path=None):
    return gen_eiscat3d_stage1("array", path=None)


@radar_generator("eiscat3d", "stage1-interp")
def gen_eiscat3d_stage1_interp(path=None, **interpolation_kwargs):
    return gen_eiscat3d_stage1("interp", path=None, **interpolation_kwargs)


@radar_generator("eiscat3d", "demo-array")
def gen_eiscat3d_demonstrator():
    """The EISCAT 3D demonstrator module."""
    tx_beam_kir = pyant.beam_of_radar("e3d_module", "array")
    rx_beam_kir = pyant.beam_of_radar("e3d_module", "array")

    kir = RX(
        lat=KIR_LAT,
        lon=KIR_LON,
        alt=KIR_ALT,
        min_elevation=30.0,
        noise=150,
        beam=rx_beam_kir,
    )
    dwell_time = 0.1
    kir_tx = TX(
        lat=KIR_LAT,
        lon=KIR_LON,
        alt=KIR_ALT,
        min_elevation=30.0,
        beam=tx_beam_kir,
        power=0.5e3 * 2 * 16,  # 500W / antenna / pol (now 16 antennas)
        bandwidth=100e3,  # 100 kHz tx bandwidth
        duty_cycle=0.25,  # 25% duty-cycle
        pulse_length=1920e-6,
        ipp=10e-3,
        n_ipp=int(dwell_time / 10e-3),
    )

    # define transmit and receive antennas for a radar network.
    tx = [kir_tx]
    rx = [kir]

    eiscat3d_demonstartor = Radar(
        tx=tx,
        rx=rx,
        min_SNRdb=10.0,
        joint_stations=[(0, 0)],
    )
    return eiscat3d_demonstartor
